from functools import wraps
from abc import ABCMeta, abstractmethod

from uniref.mono.component import *
from uniref.util.injector.interface import Injector


def _register_mono_func(func):
    @wraps(func)
    def inner(*args, **kwargs):
        return args[0]._func_set[func.__name__](*args[1:], **kwargs)
    return inner


class MonoInjector(Injector, metaclass=ABCMeta):

    @property
    def h_mono(self) -> int:
        raise NotImplementedError

    @property
    def use_il2cpp(self) -> bool:
        raise NotImplementedError

    @property
    def root_domain(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def _mono_detect(self) -> None:
        ...

    @abstractmethod
    def _build_mono_func_set(self) -> MonoNativeFuncSet:
        ...

    @abstractmethod
    def enum_assemblies(self) -> List[MonoAssembly]:
        ...

    @abstractmethod
    def guess_class_instance_address(self, klass: int, mem_writeable: bool = True) -> List[int]:
        ...

    @_register_mono_func
    def mono_get_root_domain(self) -> int:
        ...

    @_register_mono_func
    def mono_thread_attach(self, domain: int) -> int:
        ...

    @_register_mono_func
    def mono_thread_detach(self, mono_thread: int) -> None:
        ...

    @_register_mono_func
    def mono_assembly_foreach(self, func: int, user_data: int) -> int:
        ...

    @_register_mono_func
    def mono_assembly_get_image(self, assembly: int) -> int:
        ...

    @_register_mono_func
    def mono_image_get_name(self, image: int) -> str:
        ...

    @_register_mono_func
    def mono_image_get_filename(self, image: int) -> str:
        ...

    @_register_mono_func
    def mono_image_get_table_info(self, image: int, table_id: int) -> int:
        ...

    @_register_mono_func
    def mono_image_get_assembly(self, image: int) -> int:
        ...

    @_register_mono_func
    def mono_table_info_get_rows(self, table_info: int) -> int:
        ...

    @_register_mono_func
    def mono_class_get(self, image: int, token_index: int) -> int:
        ...

    @_register_mono_func
    def mono_class_get_name(self, klass: int) -> str:
        ...

    @_register_mono_func
    def mono_class_get_flags(self, klass: int) -> int:
        ...

    @_register_mono_func
    def mono_class_get_namespace(self, klass: int) -> str:
        ...

    @_register_mono_func
    def mono_class_get_fields(self, klass: int, _iter: int) -> int:
        ...

    @_register_mono_func
    def mono_class_from_mono_type(self, _type: int) -> int:
        ...

    @_register_mono_func
    def mono_class_vtable(self, domain: int, klass: int) -> int:
        ...

    @_register_mono_func
    def mono_type_get_type(self, _type: int) -> int:
        ...

    @_register_mono_func
    def mono_type_get_name(self, _type: int) -> str:
        ...

    @_register_mono_func
    def mono_type_get_class(self, _type: int) -> int:
        ...

    @_register_mono_func
    def mono_field_get_type(self, field: int) -> int:
        ...

    @_register_mono_func
    def mono_field_get_parent(self, field: int) -> int:
        ...

    @_register_mono_func
    def mono_field_get_offset(self, field: int) -> int:
        ...

    @_register_mono_func
    def mono_field_get_flags(self, _type: int) -> int:
        ...

    @_register_mono_func
    def mono_field_get_name(self, field: int) -> str:
        ...

    @_register_mono_func
    def mono_field_static_get_value(self, vtable: int, field: int, output: int) -> int:
        ...

    @_register_mono_func
    def mono_field_static_set_value(self, vtable: int, field: int, _input: int) -> int:
        ...

    @_register_mono_func
    def mono_class_get_methods(self, klass: int, _iter: int) -> int:
        ...

    @_register_mono_func
    def mono_method_get_name(self, method: int) -> str:
        ...

    @_register_mono_func
    def mono_method_get_param_names(self, method: int, names: int) -> int:
        ...

    @_register_mono_func
    def mono_method_get_class(self, method: int) -> int:
        ...

    @_register_mono_func
    def mono_compile_method(self, method: int) -> int:
        ...

    @_register_mono_func
    def mono_method_signature(self, method: int) -> int:
        ...

    @_register_mono_func
    def mono_method_get_flags(self, method: int, ptype: int) -> int:
        ...

    @_register_mono_func
    def mono_class_from_name(self, image: int, namespace: int, name: int) -> int:
        ...

    @_register_mono_func
    def mono_class_from_name_case(self, image: int, namespace: int, name: int) -> int:
        ...

    @_register_mono_func
    def mono_class_get_field_from_name(self, klass: int, name: int) -> int:
        ...

    @_register_mono_func
    def mono_class_get_method_from_name(self, klass: int, name: int, param_count: int) -> int:
        ...

    @_register_mono_func
    def mono_vtable_get_static_field_data(self, vtable: int) -> int:
        ...

    @_register_mono_func
    def mono_class_is_generic(self, klass: int) -> int:
        ...

    @_register_mono_func
    def mono_class_get_parent(self, klass: int) -> int:
        ...

    @_register_mono_func
    def mono_class_get_image(self, klass: int) -> int:
        ...

    @_register_mono_func
    def mono_class_instance_size(self, klass) -> int:
        ...

    @_register_mono_func
    def mono_domain_get(self) -> int:
        ...

    @_register_mono_func
    def mono_domain_set(self, domain: int, force: int) -> int:
        ...

    @_register_mono_func
    def mono_jit_info_table_find(self, domain: int, addr: int) -> int:
        ...

    @_register_mono_func
    def mono_jit_info_get_code_size(self, ji: int) -> int:
        ...

    @_register_mono_func
    def mono_signature_get_desc(self, signature: int, include_namespace: int) -> str:
        ...

    @_register_mono_func
    def mono_signature_get_param_count(self, signature: int) -> int:
        ...

    @_register_mono_func
    def mono_signature_get_return_type(self, signature: int) -> int:
        ...

    @_register_mono_func
    def mono_runtime_invoke(self, method: int, obj: int, params: int, exc: int):
        ...

    @_register_mono_func
    def il2cpp_domain_get_assemblies(self, domain: int, size: int) -> int:
        ...

    @_register_mono_func
    def il2cpp_image_get_class_count(self, image: int) -> int:
        ...

    @_register_mono_func
    def il2cpp_image_get_class(self, image: int, index: int) -> int:
        ...

    @_register_mono_func
    def il2cpp_method_get_param_count(self, method: int) -> int:
        ...

    @_register_mono_func
    def il2cpp_method_get_param_name(self, method: int, index: int) -> str:
        ...

    @_register_mono_func
    def il2cpp_method_get_param(self, method: int, index: int) -> int:
        ...

    @_register_mono_func
    def il2cpp_method_get_return_type(self, method: int) -> int:
        ...

    @_register_mono_func
    def il2cpp_field_static_get_value(self, field: int, output: int) -> int:
        ...

    @_register_mono_func
    def il2cpp_field_static_set_value(self, field: int, _input: int) -> int:
        ...

    def get_metadata(self) -> List[int]:
        attach = self._func_set["mono_thread_attach"].address
        detach = self._func_set["mono_thread_detach"].address
        return [self._root_domain, attach, detach]

    def _mono_init(self) -> None:
        self._func_set = self._build_mono_func_set()

        if self.use_il2cpp:
            self._root_domain = self.mono_domain_get()
        else:
            self._root_domain = self.mono_get_root_domain()

        self._func_set.set_user_data(self.get_metadata())

    def _mono_enum_assembly_impl(self, enum_assembly_callback) -> List[MonoAssembly]:
        callback_address = self.mem_alloc(alloc_size=len(enum_assembly_callback))
        self._mem_write(callback_address, enum_assembly_callback)

        user_data_address = self.mem_alloc(alloc_size=PAGE_SIZE, protection="rw-")
        self.mem_write_uint32(user_data_address, 0)

        self.mono_assembly_foreach(callback_address, user_data_address)

        cnt = self.mem_read_uint32(user_data_address)
        offset = 4 if self.bit_long == 32 else 8
        array = self.mem_read_pointer_array(user_data_address + offset, cnt)
        assemblies = [MonoAssembly(self, assembly) for assembly in array]

        self.mem_free(user_data_address)
        self.mem_free(callback_address)
        return assemblies

    def _il2cpp_enum_assembly_impl(self) -> List[MonoAssembly]:
        count = self.new_pointer(0)
        address = self.il2cpp_domain_get_assemblies(self._root_domain, count.address)
        assembly_count = count.value
        if assembly_count:
            array = self.mem_read_pointer_array(address, assembly_count)
            return [MonoAssembly(self, assembly) for assembly in array]
        return []

    def _try_get_value(self, address: int, type_name: str) -> Optional[Any]:
        try:
            if type_name == "System.Boolean":
                value = self.mem_read_bool(address)
            elif type_name == "System.SByte":
                value = self.mem_read_char(address)
            elif type_name == "System.Byte":
                value = self.mem_read_uchar(address)
            elif type_name == "System.Int16":
                value = self.mem_read_int16(address)
            elif type_name == "System.UInt16":
                value = self.mem_read_uint16(address)
            elif type_name == "System.Int32":
                value = self.mem_read_int32(address)
            elif type_name == "System.UInt32":
                value = self.mem_read_uint32(address)
            elif type_name == "System.Int64":
                value = self.mem_read_int64(address)
            elif type_name == "System.UInt64":
                value = self.mem_read_uint64(address)
            elif type_name == "System.Single":
                value = self.mem_read_float(address)
            elif type_name == "System.Double":
                value = self.mem_read_double(address)
            elif type_name == "System.String":
                value = self._read_cs_string(address)
            elif type_name == "System.Decimal":
                raise NotImplementedError("Read System.Decimal")
            else:
                value = self.mem_read_pointer(address)
        except:
            return None
        return value

    def _try_set_value(self, address: int, value: Any, type_name: str) -> bool:
        try:
            if type_name == "System.Boolean":
                self.mem_write_bool(address, value)
            elif type_name == "System.SByte":
                self.mem_write_char(address, value)
            elif type_name == "System.Byte":
                self.mem_write_uchar(address, value)
            elif type_name == "System.Int16":
                self.mem_write_int16(address, value)
            elif type_name == "System.UInt16":
                self.mem_write_uint16(address, value)
            elif type_name == "System.Int32":
                self.mem_write_int32(address, value)
            elif type_name == "System.UInt32":
                self.mem_write_uint32(address, value)
            elif type_name == "System.Int64":
                self.mem_write_int64(address, value)
            elif type_name == "System.UInt64":
                self.mem_write_uint64(address, value)
            elif type_name == "System.Single":
                self.mem_write_float(address, value)
            elif type_name == "System.Double":
                self.mem_write_double(address, value)
            elif type_name == "System.String":
                raise NotImplementedError("Modify System.String")
            elif type_name == "System.Decimal":
                raise NotImplementedError("Modify System.Decimal")
            else:
                self.mem_write_pointer(address, value)
        except TypeError:
            return False
        return True

    def native_func_exist(self, func_name: str) -> bool:
        native_func = self._func_set[func_name]
        if native_func:
            return native_func.callable()
        return False

    def get_native_func(self, func_name: str) -> Optional[MonoNativeFunc]:
        return self._func_set[func_name]

    def get_assembly_image(self, assembly: int) -> MonoImage:
        return MonoImage(self, assembly, self.mono_assembly_get_image(assembly))

    def enum_images(self) -> List[MonoImage]:
        images = []
        assemblies = self.enum_assemblies()
        for assembly in assemblies:
            handle = assembly.handle
            image = self.mono_assembly_get_image(handle) & 0xFFFFFFFFFFFF
            images.append(MonoImage(self, handle, image))
        return images

    def enum_classes_in_image(self, image: int) -> List[MonoClass]:
        classes = []
        if self.use_il2cpp:
            class_count = 0
            if self.native_func_exist("il2cpp_image_get_class_count") \
                    and self.native_func_exist("il2cpp_image_get_class"):
                class_count = self.il2cpp_image_get_class_count(image)
            for i in range(class_count):
                klass = self.il2cpp_image_get_class(image, i) & 0xFFFFFFFFFFFF
                if klass:
                    classes.append(MonoClass(self, image, klass))
        else:
            type_define = self.mono_image_get_table_info(image, MONO_TABLE_TYPEDEF)
            if not type_define:
                return []
            define_cnt = self.mono_table_info_get_rows(type_define)
            for i in range(define_cnt):
                klass = self.mono_class_get(image, MONO_TOKEN_TYPE_DEF | i + 1)
                if klass:
                    classes.append(MonoClass(self, image, klass))
        return classes

    def enum_fields_in_class(self, klass: int) -> List[MonoField]:
        fields = []
        iter_ptr = self.new_pointer(0)
        while True:
            field = self.mono_class_get_fields(klass, iter_ptr.address) & 0xFFFFFFFFFFFF
            if not field:
                break
            fields.append(MonoField(self, klass, field))
        return fields

    def enum_methods_in_class(self, klass: int) -> List[MonoMethod]:
        methods = []
        iter_ptr = self.new_pointer(0)
        while True:
            method = self.mono_class_get_methods(klass, iter_ptr.address) & 0xFFFFFFFFFFFF
            if not method:
                break
            methods.append(MonoMethod(self, klass, method))
        return methods

    def find_image_by_name(self, image_name: str) -> Optional[MonoImage]:
        assemblies = self.enum_assemblies()
        for assembly in assemblies:
            image = self.mono_assembly_get_image(assembly.handle) & 0xFFFFFFFFFFFF
            name = self.mono_image_get_name(image)
            if name == image_name:
                return MonoImage(self, assembly.handle, image, image_name=name)
        return None

    def get_image_assembly(self, image: int) -> MonoAssembly:
        assembly = self.mono_image_get_assembly(image) & 0xFFFFFFFFFFFF
        return MonoAssembly(self, assembly) if assembly else None

    def get_image_name(self, image: int) -> str:
        return self.mono_image_get_name(image)

    def get_image_filename(self, image: int) -> str:
        return self.mono_image_get_filename(image)

    def find_class_in_image(self, image: int, namespace: str, name: str) -> Optional[MonoClass]:
        name = self.new_c_string(name)
        namespace = self.new_c_string(namespace)

        klass = 0
        if self.native_func_exist("mono_class_from_name_case"):
            klass = self.mono_class_from_name_case(image, namespace.address, name.address)
        if not klass:
            klass = self.mono_class_from_name(image, namespace.address, name.address)
        klass &= 0xFFFFFFFFFFFF
        return MonoClass(self, image, klass) if klass else None

    def get_class_name(self, klass: int) -> str:
        return self.mono_class_get_name(klass)

    def get_class_namespace(self, klass: int) -> str:
        return self.mono_class_get_namespace(klass)

    def get_parent_class(self, klass: int) -> Optional[MonoClass]:
        parent = self.mono_class_get_parent(klass) & 0xFFFFFFFFFFFF
        return MonoClass(self, None, parent) if parent else None

    def get_class_image(self, klass: int) -> Optional[MonoImage]:
        image = self.mono_class_get_image(klass) & 0xFFFFFFFFFFFF
        return MonoImage(self, None, image) if image else None

    def get_class_vtable(self, klass: int) -> int:
        if self.use_il2cpp:
            return klass
        else:
            return self.mono_class_vtable(self._root_domain, klass)

    def find_field_in_class(self, klass: int, name: str) -> Optional[MonoField]:
        name = self.new_c_string(name)
        field = self.mono_class_get_field_from_name(klass, name.address) & 0xFFFFFFFFFFFF
        return MonoField(self, klass, field) if field else None

    def get_static_field_address(self, klass: int, field: int) -> int:
        if not self.use_il2cpp:
            vtable = self.get_class_vtable(klass)
            if vtable:
                static_field_data = self.mono_vtable_get_static_field_data(vtable)
                if static_field_data > 0x10000:
                    return static_field_data + self.mono_field_get_offset(field)
        return 0

    def get_static_field_value(self, klass: int, field: int, type_name: str) -> Optional[Any]:
        if self.use_il2cpp:
            output = self.new_uint64(0)
            self.il2cpp_field_static_get_value(field, output.address)
            return self._try_get_value(output.address, type_name)
        else:
            address = self.get_static_field_address(klass, field)
            if not address:
                raise AttributeError("Static field address not found")
            return self._try_get_value(address, type_name)

    def set_static_field_value(self, klass: int, field: int, type_name: str, value: Any) -> bool:
        if self.use_il2cpp:
            _input = self.new_uint64(0)
            if self._try_set_value(_input.address, value, type_name):
                self.il2cpp_field_static_set_value(field, _input.address)
                return True
            return False
        else:
            address = self.get_static_field_address(klass, field)
            if not address:
                raise AttributeError("Static field address not found")
            return self._try_set_value(address, value, type_name)

    def find_method_in_class(self, klass: int, name: str, param_count: int = -1) -> Optional[MonoMethod]:
        name = self.new_c_string(name)
        method = self.mono_class_get_method_from_name(klass, name.address, param_count) & 0xFFFFFFFFFFFF
        return MonoMethod(self, klass, method) if method else None

    def get_method_name(self, method: int) -> str:
        return self.mono_method_get_name(method)

    def get_method_size(self, method: int) -> int:
        if self.native_func_exist("mono_jit_info_table_find"):
            info_table = self.mono_jit_info_table_find(self._root_domain, method)
            if info_table:
                return self.mono_jit_info_get_code_size(info_table)
        return -1

    def get_method_class(self, method: int) -> Optional[MonoClass]:
        klass = self.mono_method_get_class(method) & 0xFFFFFFFFFFFF
        return MonoClass(self, None, klass) if klass else None

    def get_method_signature(self, method: int) -> str:
        if self.use_il2cpp:
            param_count = self.il2cpp_method_get_param_count(method)
            param_info = ''
            for i in range(param_count):
                param_name = self.il2cpp_method_get_param_name(method, i)
                param_type = self.il2cpp_method_get_param(method, i)
                type_name = self.mono_type_get_name(param_type)
                param_info += f"{type_name} {param_name}, "
            return_type = self.il2cpp_method_get_return_type(method)
            return_type_str = self.mono_type_get_name(return_type)
            return f"{return_type_str} ({param_info[:-2]})"
        else:
            method_sig = self.mono_method_signature(method)
            param_desc = self.mono_signature_get_desc(method_sig, 1)
            param_count = self.mono_signature_get_param_count(method_sig)

            if param_count:
                names_ptr_array = self.new_pointer_array([0 for _ in range(param_count)])
                self.mono_method_get_param_names(method, names_ptr_array.address)

                param_names = []
                names_ptr_array = names_ptr_array.elements
                for i in range(param_count):
                    param_names.append(self.mem_read_c_string(names_ptr_array[i]))

            return_type = self.mono_signature_get_return_type(method_sig)
            return_type_str = self.mono_type_get_name(return_type)

            param_info = ''
            if param_count or param_desc:
                param_types = param_desc.split(',')
                if param_count == len(param_types):
                    param_info = ', '.join([f"{param_types[i]} {param_names[i]}" for i in range(param_count)])
                else:
                    param_info = "<parse error>"

            return f"{return_type_str} ({param_info})"

    def compile_method(self, method: int) -> int:
        if self.use_il2cpp:
            return self.mem_read_pointer(method)
        else:
            klass = self.mono_method_get_class(method)
            if not klass:
                return 0
            is_generic_exist = self.native_func_exist("mono_class_is_generic")
            if (is_generic_exist and self.mono_class_is_generic(klass) == 0) or (not is_generic_exist):
                return self.mono_compile_method(method)
        return 0



from uniref.mono.component import *
from uniref.util.winapi import WinApi
from uniref.util.scanner import CMemoryScanner
from uniref.util.injector.windows import WinInjector
from uniref.mono.injector.interface import MonoInjector


class WinMonoInjector(WinInjector, MonoInjector):
    """ Unity application injector for ``Windows``. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._h_mono = 0
        self._func_set = None
        self._root_domain = 0
        self._use_il2cpp = False

        self._mono_detect()
        self._mono_init()

    @property
    def h_mono(self) -> int:
        return self._h_mono

    @property
    def use_il2cpp(self) -> bool:
        return self._use_il2cpp

    @property
    def root_domain(self) -> int:
        return self._root_domain

    def _mono_detect(self) -> None:
        h_mono = self.get_module_base("mono.dll") or self.get_module_base("mono-2.0-bdwgc.dll")
        if h_mono:
            self._h_mono = h_mono
        else:
            modules = WinApi.GetRemoteModules(self._h_process)
            for module in modules:
                if self.get_proc_address(module, "il2cpp_thread_attach"):
                    self._h_mono = module
                    self._use_il2cpp = True
                    break
                if self.get_proc_address(module, "mono_thread_attach"):
                    self._h_mono = module
                    break
            if self._h_mono == 0:
                raise SystemError("Only support mono & il2cpp application")

    def _find_mono_api_address(self, page_start: int, proc_name: str) -> int:
        return_address = page_start + 0x300
        proc_name_address = page_start + 0x200

        if self._bit_long == 32:
            code = f"push {hex(proc_name_address)}           \n" \
                   f"push {hex(self._h_mono)}                \n" \
                   f"mov eax, {hex(self._get_proc_address)}  \n" \
                   f"call eax                                \n" \
                   f"mov ecx, {hex(return_address)}          \n" \
                   f"mov dword ptr [ecx], eax                \n" \
                   f"ret"
        else:
            code = f"sub rsp, 28h                            \n" \
                   f"mov rcx, {hex(self._h_mono)}            \n" \
                   f"mov rdx, {hex(proc_name_address)}       \n" \
                   f"mov rax, {hex(self._get_proc_address)}  \n" \
                   f"call rax                                \n" \
                   f"mov r12, {hex(return_address)}          \n" \
                   f"mov qword ptr [r12], rax                \n" \
                   f"add rsp, 28h                            \n" \
                   f"ret"

        self._mem_write(proc_name_address, proc_name.encode() + b'\x00')
        self.code_execute(code, page_start)
        return self.mem_read_pointer(return_address)

    def _build_mono_func_set(self) -> MonoNativeFuncSet:
        func_set = MonoNativeFuncSet(self._use_il2cpp)
        page_start = self.mem_alloc()
        if self._use_il2cpp:
            for fn in il2cpp_native_func_name:
                mono_name = il2cpp_mono_native_func_map[fn]
                func_set[mono_name] = self._find_mono_api_address(page_start, fn)
                func_set[mono_name].set_mono_injector(self)
        else:
            for fn in mono_native_func_name:
                func_set[fn] = self._find_mono_api_address(page_start, fn)
                func_set[fn].set_mono_injector(self)
        self.mem_free(page_start)
        return func_set

    def enum_assemblies(self) -> List[MonoAssembly]:
        if self._use_il2cpp:
            return self._il2cpp_enum_assembly_impl()
        else:
            if self._bit_long == 32:
                enum_assembly_callback = b"\x55\x8b\xec\x8b\x4d\x0c\x8b\x11\x81\xfa\xfe\x03\x00\x00\x77\x09\x8b\x45\x08\x89\x44\x91\x04\xff\x01\x5d\xc3"
                """
                    void _cdecl enum_assembly_callback(void* domain, CUSTOM_DOMAIN_ARRAY32* v)
                    {
                        if (v->cnt <= 1022) {
                            v->domains[v->cnt] = (UINT32)domain;
                            v->cnt += 1;
                        }
                    }
                """
            else:
                enum_assembly_callback = b"\x8b\x02\x3d\xfe\x01\x00\x00\x77\x07\x48\x89\x4c\xc2\x08\xff\x02\xc3"
                """
                    void _cdecl enum_assembly_callback(void* domain, CUSTOM_DOMAIN_ARRAY64* v)
                    {
                        if (v->cnt <= 510) {
                            v->domains[v->cnt] = (UINT64)domain;
                            v->cnt += 1;
                        }
                    }
                """
            return self._mono_enum_assembly_impl(enum_assembly_callback)

    def guess_class_instance_address(self, klass: int, mem_writeable: bool = False) -> List[int]:
        vtable = self.get_class_vtable(klass)
        if vtable:
            scanner = CMemoryScanner(self._process_id, self._bit_long)
            return scanner.scan_pointer(vtable, mem_writeable)
        return []



from keystone import *
from abc import ABCMeta


class _Compiler(metaclass=ABCMeta):

    @property
    def compiler(self):
        raise NotImplementedError

    def compile(self, code: str, address: int = 0) -> bytes:
        if not isinstance(code, str):
            raise TypeError("code should be str")
        encoding, _ = self.compiler.asm(code, address)
        return bytes(encoding)


class X86Compiler(_Compiler):

    def __init__(self, bit_long: int = 64, ATT_syntax: bool = False) -> None:
        if bit_long == 32:
            ks_mode = KS_MODE_32
        elif bit_long == 64:
            ks_mode = KS_MODE_64
        else:
            raise NotImplementedError("Current architecture is not supported")
        self._compiler = Ks(KS_ARCH_X86, ks_mode)
        if ATT_syntax:
            self._compiler.syntax = KS_OPT_SYNTAX_ATT

    @property
    def compiler(self):
        return self._compiler


class ArmCompiler(_Compiler):

    def __init__(self, bit_long: int = 64, thumb: bool = False) -> None:
        if bit_long == 32:
            self._compiler = Ks(KS_ARCH_ARM, KS_MODE_THUMB if thumb else KS_MODE_ARM)
        elif bit_long == 64:
            self._compiler = Ks(KS_ARCH_ARM64, KS_MODE_LITTLE_ENDIAN)
        else:
            raise NotImplementedError("Current architecture is not supported")

    @property
    def compiler(self):
        return self._compiler



from uniref import WinUniRef

ref = WinUniRef("Lethal Company.exe")

manager = ref.find_class_in_image("Assembly-CSharp", "GameNetworkManager")
manager.instance = manager.find_field("<Instance>k__BackingField").value

manager.find_field("maxAllowedPlayers").value = 10

terminal = ref.find_class_in_image("Assembly-CSharp", "Terminal")
money = terminal.find_field("groupCredits")

for address in terminal.guess_instance_address():
    money.instance = address
    if money.value == 60:
        money.value = 999999999



INFINITE = -1
INVALID_HANDLE_VALUE = -1

LIST_MODULES_DEFAULT = 0x00
LIST_MODULES_32BIT = 0x01
LIST_MODULES_64BIT = 0x02
LIST_MODULES_ALL = 0x03

TH32CS_SNAPHEAPLIST = 0x00000001
TH32CS_SNAPPROCESS = 0x00000002
TH32CS_SNAPTHREAD = 0x00000004
TH32CS_SNAPMODULE = 0x00000008
TH32CS_SNAPMODULE32 = 0x00000010
TH32CS_SNAPALL = (TH32CS_SNAPHEAPLIST | TH32CS_SNAPPROCESS | TH32CS_SNAPTHREAD | TH32CS_SNAPMODULE)
TH32CS_INHERIT = 0x80000000

SYNCHRONIZE = 0x00100000
STANDARD_RIGHTS_REQUIRED = 0x000F0000

PROCESS_TERMINATE = 0x0001
PROCESS_CREATE_THREAD = 0x0002
PROCESS_SET_SESSIONID = 0x0004
PROCESS_VM_OPERATION = 0x0008
PROCESS_VM_READ = 0x0010
PROCESS_VM_WRITE = 0x0020
PROCESS_DUP_HANDLE = 0x0040
PROCESS_CREATE_PROCESS = 0x0080
PROCESS_SET_QUOTA = 0x0100
PROCESS_SET_INFORMATION = 0x0200
PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_SUSPEND_RESUME = 0x0800
PROCESS_QUERY_LIMITED_INFORMATION = 0x1000
PROCESS_SET_LIMITED_INFORMATION = 0x2000

# NTDDI_VERSION >= NTDDI_VISTA
PROCESS_ALL_ACCESS = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0xFFFF

# NTDDI_VERSION < NTDDI_VISTA
# PROCESS_ALL_ACCESS                 = STANDARD_RIGHTS_REQUIRED | SYNCHRONIZE | 0xFFF

PAGE_SIZE = 4096
PAGE_NOACCESS = 0x01
PAGE_READONLY = 0x02
PAGE_READWRITE = 0x04
PAGE_WRITECOPY = 0x08
PAGE_EXECUTE = 0x10
PAGE_EXECUTE_READ = 0x20
PAGE_EXECUTE_READWRITE = 0x40
PAGE_EXECUTE_WRITECOPY = 0x80
PAGE_GUARD = 0x100
PAGE_NOCACHE = 0x200
PAGE_WRITECOMBINE = 0x400
PAGE_GRAPHICS_NOACCESS = 0x0800
PAGE_GRAPHICS_READONLY = 0x1000
PAGE_GRAPHICS_READWRITE = 0x2000
PAGE_GRAPHICS_EXECUTE = 0x4000
PAGE_GRAPHICS_EXECUTE_READ = 0x8000
PAGE_GRAPHICS_EXECUTE_READWRITE = 0x10000
PAGE_GRAPHICS_COHERENT = 0x20000
PAGE_GRAPHICS_NOCACHE = 0x40000
PAGE_ENCLAVE_THREAD_CONTROL = 0x80000000
PAGE_REVERT_TO_FILE_MAP = 0x80000000
PAGE_TARGETS_NO_UPDATE = 0x40000000
PAGE_TARGETS_INVALID = 0x40000000
PAGE_ENCLAVE_UNVALIDATED = 0x20000000
PAGE_ENCLAVE_MASK = 0x10000000
PAGE_ENCLAVE_DECOMMIT = (PAGE_ENCLAVE_MASK | 0)
PAGE_ENCLAVE_SS_FIRST = (PAGE_ENCLAVE_MASK | 1)
PAGE_ENCLAVE_SS_REST = (PAGE_ENCLAVE_MASK | 2)
MEM_COMMIT = 0x00001000
MEM_RESERVE = 0x00002000
MEM_REPLACE_PLACEHOLDER = 0x00004000
MEM_RESERVE_PLACEHOLDER = 0x00040000
MEM_RESET = 0x00080000
MEM_TOP_DOWN = 0x00100000
MEM_WRITE_WATCH = 0x00200000
MEM_PHYSICAL = 0x00400000
MEM_ROTATE = 0x00800000
MEM_DIFFERENT_IMAGE_BASE_OK = 0x00800000
MEM_RESET_UNDO = 0x01000000
MEM_LARGE_PAGES = 0x20000000
MEM_4MB_PAGES = 0x80000000
MEM_64K_PAGES = (MEM_LARGE_PAGES | MEM_PHYSICAL)
MEM_UNMAP_WITH_TRANSIENT_BOOST = 0x00000001
MEM_COALESCE_PLACEHOLDERS = 0x00000001
MEM_PRESERVE_PLACEHOLDER = 0x00000002
MEM_DECOMMIT = 0x00004000
MEM_RELEASE = 0x00008000
MEM_FREE = 0x00010000

TYPE_VOID = 0
TYPE_BOOL = 1
TYPE_CHAR = 2
TYPE_UCHAR = 3
TYPE_INT16 = 4
TYPE_UINT16 = 5
TYPE_INT32 = 6
TYPE_UINT32 = 7
TYPE_INT64 = 8
TYPE_UINT64 = 9
TYPE_FLOAT = 10
TYPE_DOUBLE = 11
TYPE_VOID_P = 12
TYPE_CHAR_P = 13
TYPE_CS_STRING = 14


# enum MonoMetaTableEnum
MONO_TABLE_MODULE = 0
MONO_TABLE_TYPEREF = 1
MONO_TABLE_TYPEDEF = 2
MONO_TABLE_FIELD_POINTER = 3
MONO_TABLE_FIELD = 4
MONO_TABLE_METHOD_POINTER = 5
MONO_TABLE_METHOD = 6
MONO_TABLE_PARAM_POINTER = 7
MONO_TABLE_PARAM = 8
MONO_TABLE_INTERFACEIMPL = 9
MONO_TABLE_MEMBERREF = 10
MONO_TABLE_CONSTANT = 11
MONO_TABLE_CUSTOMATTRIBUTE = 12
MONO_TABLE_FIELDMARSHAL = 13
MONO_TABLE_DECLSECURITY = 14
MONO_TABLE_CLASSLAYOUT = 15
MONO_TABLE_FIELDLAYOUT = 16
MONO_TABLE_STANDALONESIG = 17
MONO_TABLE_EVENTMAP = 18
MONO_TABLE_EVENT_POINTER = 19
MONO_TABLE_EVENT = 20
MONO_TABLE_PROPERTYMAP = 21
MONO_TABLE_PROPERTY_POINTER = 22
MONO_TABLE_PROPERTY = 23
MONO_TABLE_METHODSEMANTICS = 24
MONO_TABLE_METHODIMPL = 25
MONO_TABLE_MODULEREF = 26
MONO_TABLE_TYPESPEC = 27
MONO_TABLE_IMPLMAP = 28
MONO_TABLE_FIELDRVA = 29
MONO_TABLE_UNUSED6 = 30
MONO_TABLE_UNUSED7 = 31
MONO_TABLE_ASSEMBLY = 32
MONO_TABLE_ASSEMBLYPROCESSOR = 33
MONO_TABLE_ASSEMBLYOS = 34
MONO_TABLE_ASSEMBLYREF = 35
MONO_TABLE_ASSEMBLYREFPROCESSOR = 36
MONO_TABLE_ASSEMBLYREFOS = 37
MONO_TABLE_FILE = 38
MONO_TABLE_EXPORTEDTYPE = 39
MONO_TABLE_MANIFESTRESOURCE = 40
MONO_TABLE_NESTEDCLASS = 41
MONO_TABLE_GENERICPARAM = 42
MONO_TABLE_METHODSPEC = 43
MONO_TABLE_GENERICPARAMCONSTRAINT = 44

# enum MonoTokenType
MONO_TOKEN_MODULE = 0x00000000
MONO_TOKEN_TYPE_REF = 0x01000000
MONO_TOKEN_TYPE_DEF = 0x02000000
MONO_TOKEN_FIELD_DEF = 0x04000000
MONO_TOKEN_METHOD_DEF = 0x06000000
MONO_TOKEN_PARAM_DEF = 0x08000000
MONO_TOKEN_INTERFACE_IMPL = 0x09000000
MONO_TOKEN_MEMBER_REF = 0x0a000000
MONO_TOKEN_CUSTOM_ATTRIBUTE = 0x0c000000
MONO_TOKEN_PERMISSION = 0x0e000000
MONO_TOKEN_SIGNATURE = 0x11000000
MONO_TOKEN_EVENT = 0x14000000
MONO_TOKEN_PROPERTY = 0x17000000
MONO_TOKEN_MODULE_REF = 0x1a000000
MONO_TOKEN_TYPE_SPEC = 0x1b000000
MONO_TOKEN_ASSEMBLY = 0x20000000
MONO_TOKEN_ASSEMBLY_REF = 0x23000000
MONO_TOKEN_FILE = 0x26000000
MONO_TOKEN_EXPORTED_TYPE = 0x27000000
MONO_TOKEN_MANIFEST_RESOURCE = 0x28000000
MONO_TOKEN_GENERIC_PARAM = 0x2a000000
MONO_TOKEN_METHOD_SPEC = 0x2b000000
MONO_TOKEN_STRING = 0x70000000
MONO_TOKEN_NAME = 0x71000000
MONO_TOKEN_BASE_TYPE = 0x72000000

FIELD_ATTRIBUTE_FIELD_ACCESS_MASK = 0x0007
FIELD_ATTRIBUTE_COMPILER_CONTROLLED = 0x0000
FIELD_ATTRIBUTE_PRIVATE = 0x0001
FIELD_ATTRIBUTE_FAM_AND_ASSEM = 0x0002
FIELD_ATTRIBUTE_ASSEMBLY = 0x0003
FIELD_ATTRIBUTE_FAMILY = 0x0004
FIELD_ATTRIBUTE_FAM_OR_ASSEM = 0x0005
FIELD_ATTRIBUTE_PUBLIC = 0x0006
FIELD_ATTRIBUTE_STATIC = 0x0010
FIELD_ATTRIBUTE_INIT_ONLY = 0x0020
FIELD_ATTRIBUTE_LITERAL = 0x0040
FIELD_ATTRIBUTE_NOT_SERIALIZED = 0x0080
FIELD_ATTRIBUTE_SPECIAL_NAME = 0x0200
FIELD_ATTRIBUTE_PINVOKE_IMPL = 0x2000
FIELD_ATTRIBUTE_RESERVED_MASK = 0x9500
FIELD_ATTRIBUTE_RT_SPECIAL_NAME = 0x0400
FIELD_ATTRIBUTE_HAS_FIELD_MARSHAL = 0x1000
FIELD_ATTRIBUTE_HAS_DEFAULT = 0x8000
FIELD_ATTRIBUTE_HAS_FIELD_RVA = 0x0100

METHOD_ATTRIBUTE_MEMBER_ACCESS_MASK = 0x0007
METHOD_ATTRIBUTE_COMPILER_CONTROLLED = 0x0000
METHOD_ATTRIBUTE_PRIVATE = 0x0001
METHOD_ATTRIBUTE_FAM_AND_ASSEM = 0x0002
METHOD_ATTRIBUTE_ASSEM = 0x0003
METHOD_ATTRIBUTE_FAMILY = 0x0004
METHOD_ATTRIBUTE_FAM_OR_ASSEM = 0x0005
METHOD_ATTRIBUTE_PUBLIC = 0x0006

METHOD_ATTRIBUTE_STATIC = 0x0010
METHOD_ATTRIBUTE_FINAL = 0x0020
METHOD_ATTRIBUTE_VIRTUAL = 0x0040
METHOD_ATTRIBUTE_HIDE_BY_SIG = 0x0080

METHOD_ATTRIBUTE_VTABLE_LAYOUT_MASK = 0x0100
METHOD_ATTRIBUTE_REUSE_SLOT = 0x0000
METHOD_ATTRIBUTE_NEW_SLOT = 0x0100

METHOD_ATTRIBUTE_STRICT = 0x0200
METHOD_ATTRIBUTE_ABSTRACT = 0x0400
METHOD_ATTRIBUTE_SPECIAL_NAME = 0x0800

METHOD_ATTRIBUTE_PINVOKE_IMPL = 0x2000
METHOD_ATTRIBUTE_UNMANAGED_EXPORT = 0x0008

MONO_TYPE_NAME_FORMAT_IL = 0
MONO_TYPE_NAME_FORMAT_REFLECTION = 1
MONO_TYPE_NAME_FORMAT_FULL_NAME = 2
MONO_TYPE_NAME_FORMAT_ASSEMBLY_QUALIFIED = 3


CALL_TYPE_DEF = 0
CALL_TYPE_CDECL = 1
CALL_TYPE_STDCALL = 2
CALL_TYPE_THISCALL = 3
CALL_TYPE_FASTCALL = 4



from typing import *
from capstone import *


class X86Disasm:

    def __init__(self, bit_long: int = 64, ATT_syntax: bool = False) -> None:
        if bit_long == 32:
            cs_mode = CS_MODE_32
        elif bit_long == 64:
            cs_mode = CS_MODE_64
        else:
            raise NotImplementedError("Current architecture is not supported")
        self._disasm = Cs(CS_ARCH_X86, cs_mode)
        self._disasm.detail = True
        if ATT_syntax:
            self._disasm.syntax = CS_OPT_SYNTAX_ATT

    def disassemble(self, code: bytes, base_address: int = 0, count: int = 0) -> List[CsInsn]:
        return [ins for ins in self._disasm.disasm(code, base_address, count)]



from typing import *


class _BasicType:
    """ Base class of ``Bool``, ``Char``, ``UChar``, ``Int16``, ``UInt16``, ``Int32``, ``UInt32``,
    ``Int64``, ``UInt64``, ``Float``, ``Double`` and ``Pointer``.
    
    Examples:

    >>> from uniref import WinUniRef
    >>> ref = WinUniRef("game.exe") 
    >>> int32 = ref.injector.new_int32(1) 
    >>> type(int32) 
    <class 'uniref.define.types.Int32'>
    >>> hex(int32.address)
    '0xd00000'
    >>> int32.value
    1
    >>> int32.value = 12345
    >>> int32.value
    12345
    >>> int32.release()

    """
    def __init__(self, injector: object, value: Any, elem_size: int, auto_release: bool = True) -> None:
        self._injector = injector
        if elem_size <= 0:
            raise ValueError("Element size should be positive")
        self._value = value
        self._elem_size = elem_size
        self._auto_release = auto_release

        self._address = self._injector.mem_alloc(alloc_size=elem_size, protection="rw-")
        self._assign_value()

    def __del__(self):
        if self._auto_release:
            self.release()

    @property
    def value(self) -> int:
        """ variable value """
        if self._address == 0:
            raise MemoryError(f"Memory has been freed")
        return self._do_read(self._address)

    @value.setter
    def value(self, value: Any) -> None:
        self.set_value(value)

    def set_value(self, value: Any) -> None:
        """ Set variable value. """
        self._value = value
        self._assign_value()

    @property
    def address(self) -> int:
        """ variable address """
        return self._address

    @property
    def auto_release(self) -> bool:
        """ auto release """
        return self._auto_release

    @auto_release.setter
    def auto_release(self, value: bool) -> None:
        self.set_auto_release(value)

    def set_auto_release(self, value: bool) -> None:
        """ Set whether to release automatically. """
        if not isinstance(value, bool):
            raise TypeError("auto_release should be bool")
        self._auto_release = value

    def _assign_value(self) -> None:
        if self._address == 0:
            raise MemoryError(f"Memory has been freed")
        self._do_write(self._address, self._value)

    def release(self) -> None:
        """ Release the memory. """
        if self._address:
            self._injector.mem_free(self._address)
            self._address = 0


class Bool(_BasicType):

    def __init__(self, injector: object, value: bool, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_bool
        self._do_write = injector.mem_write_bool
        super(Bool, self).__init__(injector, value, 1, auto_release)


class Char(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_char
        self._do_write = injector.mem_write_char
        super(Char, self).__init__(injector, value, 1, auto_release)


class UChar(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_uchar
        self._do_write = injector.mem_write_uchar
        super(UChar, self).__init__(injector, value, 1, auto_release)


class Int16(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_int16
        self._do_write = injector.mem_write_int16
        super(Int16, self).__init__(injector, value, 2, auto_release)


class UInt16(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_uint16
        self._do_write = injector.mem_write_uint16
        super(UInt16, self).__init__(injector, value, 2, auto_release)


class Int32(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_int32
        self._do_write = injector.mem_write_int32
        super(Int32, self).__init__(injector, value, 4, auto_release)


class UInt32(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_uint32
        self._do_write = injector.mem_write_uint32
        super(UInt32, self).__init__(injector, value, 4, auto_release)


class Int64(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_int64
        self._do_write = injector.mem_write_int64
        super(Int64, self).__init__(injector, value, 8, auto_release)


class UInt64(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_uint64
        self._do_write = injector.mem_write_uint64
        super(UInt64, self).__init__(injector, value, 8, auto_release)


class Float(_BasicType):

    def __init__(self, injector: object, value: float, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_float
        self._do_write = injector.mem_write_float
        super(Float, self).__init__(injector, value, 4, auto_release)


class Double(_BasicType):

    def __init__(self, injector: object, value: float, auto_release: bool = True) -> None:
        self._do_read = injector.mem_read_double
        self._do_write = injector.mem_write_double
        super(Double, self).__init__(injector, value, 8, auto_release)


class Pointer(_BasicType):

    def __init__(self, injector: object, value: int, auto_release: bool = True):
        self._do_read = injector.mem_read_pointer
        self._do_write = injector.mem_write_pointer
        elem_size = 4 if injector.bit_long == 32 else 8
        super(Pointer, self).__init__(injector, value, elem_size, auto_release)


class _ArrayType:
    """ Base class of ``BoolArray``, ``CharArray``, ``UCharArray``, ``Int16Array``, ``UInt16Array``,
    ``Int32Array``, ``UInt32Array``, ``Int64Array``, ``UInt64Array``, ``FloatArray``, ``DoubleArray`` and ``PointerArray``.

    Examples:

    >>> from uniref import WinUniRef
    >>> ref = WinUniRef("game.exe") 
    >>> array = ref.injector.new_double_array([1.5, 5.1, 9.2])
    >>> type(array) 
    <class 'uniref.define.types.DoubleArray'>
    >>> hex(array.address)
    '0xe00000'
    >>> array.elem_count
    3
    >>> array.elements
    [1.5, 5.1, 9.2]
    >>> array[0]
    1.5
    >>> array[0] = 2.5
    >>> array.elements
    [2.5, 5.1, 9.2]
    >>> array.elements = [0.1, 0.2, 0.3]
    >>> array.elements
    [0.1, 0.2, 0.3]
    >>> array.release()

    """
    def __init__(
            self,
            injector: object,
            elem_size: int,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[Any]] = None,
            auto_release: bool = True
    ) -> None:
        self._injector = injector
        if elem_size <= 0:
            raise ValueError("Element size should be positive")
        self._elem_size = elem_size

        if elements:
            elem_count = len(elements)
        else:
            elem_count = element_count
        if elem_count <= 0:
            raise ValueError("Array is empty")
        if address > 0:
            self._address = address
        else:
            self._address = self._injector.mem_alloc(alloc_size=elem_size * elem_count, protection="rw-")

        self._elements = elements
        self._elem_count = elem_count
        self._auto_release = auto_release

        if elements:
            self._assign_array()

    def __del__(self):
        if self._auto_release:
            self.release()

    def __getitem__(self, idx: int) -> Any:
        if not isinstance(idx, int):
            raise TypeError("Use int as array index")
        if idx >= self._elem_count:
            raise IndexError(f"Index is greater than array size")
        if self._address == 0:
            raise MemoryError(f"Memory has been freed")
        return self._do_read(self._address + self._elem_size * idx)

    def __setitem__(self, idx: int, value: Any) -> None:
        if not isinstance(idx, int):
            raise TypeError("Use int as array index")
        if idx >= self._elem_count:
            raise IndexError(f"Index is greater than array size")
        if self._address == 0:
            raise MemoryError(f"Memory has been freed")
        self._do_write(self._address + self._elem_size * idx, value)

    @property
    def address(self) -> int:
        """ array address """
        return self._address

    @property
    def elements(self) -> List[Any]:
        """ array elements """
        if self._address == 0:
            raise MemoryError(f"Memory has been freed")
        elements = []
        for i in range(self._elem_count):
            elements.append(self._do_read(self._address + self._elem_size * i))
        return elements

    @elements.setter
    def elements(self, value: List[Any]) -> None:
        self.set_elements(value)

    def set_elements(self, value: List[Any]) -> None:
        """ Update array elements. """
        if not isinstance(value, list):
            raise TypeError("Elements should be list")
        if len(value) != self._elem_count:
            raise ValueError("Size of the new array is different from the original length")
        self._elements = value
        self._assign_array()

    @property
    def elem_size(self) -> int:
        return self._elem_size

    @property
    def elem_count(self) -> int:
        """ number of array elements """
        return self._elem_count

    @property
    def auto_release(self) -> bool:
        """ auto release """
        return self._auto_release

    @auto_release.setter
    def auto_release(self, value: bool) -> None:
        self.set_auto_release(value)

    def set_auto_release(self, value: bool) -> None:
        """ Set whether to release automatically. """
        if not isinstance(value, bool):
            raise TypeError("auto_release should be bool")
        self._auto_release = value

    def _assign_array(self) -> None:
        if self._address == 0:
            raise MemoryError(f"Memory has been freed")
        for i in range(self._elem_count):
            self._do_write(self._address + self._elem_size * i, self._elements[i])

    def release(self) -> None:
        """ Release the memory. """
        if self._address:
            self._injector.mem_free(self._address)
            self._address = 0


class BoolArray(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[bool]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_bool
        self._do_write = injector.mem_write_bool
        super(BoolArray, self).__init__(injector, 1, address, element_count, elements, auto_release)


class CharArray(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_char
        self._do_write = injector.mem_write_char
        super(CharArray, self).__init__(injector, 1, address, element_count, elements, auto_release)


class UCharArray(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_uchar
        self._do_write = injector.mem_write_uchar
        super(UCharArray, self).__init__(injector, 1, address, element_count, elements, auto_release)


class Int16Array(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_int16
        self._do_write = injector.mem_write_int16
        super(Int16Array, self).__init__(injector, 2, address, element_count, elements, auto_release)


class UInt16Array(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_uint16
        self._do_write = injector.mem_write_uint16
        super(UInt16Array, self).__init__(injector, 2, address, element_count, elements, auto_release)


class Int32Array(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_int32
        self._do_write = injector.mem_write_int32
        super(Int32Array, self).__init__(injector, 4, address, element_count, elements, auto_release)


class UInt32Array(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_uint32
        self._do_write = injector.mem_write_uint32
        super(UInt32Array, self).__init__(injector, 4, address, element_count, elements, auto_release)


class Int64Array(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_int64
        self._do_write = injector.mem_write_int64
        super(Int64Array, self).__init__(injector, 8, address, element_count, elements, auto_release)


class UInt64Array(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_uint64
        self._do_write = injector.mem_write_uint64
        super(UInt64Array, self).__init__(injector, 8, address, element_count, elements, auto_release)


class FloatArray(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[float]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_float
        self._do_write = injector.mem_write_float
        super(FloatArray, self).__init__(injector, 4, address, element_count, elements, auto_release)


class DoubleArray(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[float]] = None,
            auto_release: bool = True
    ) -> None:
        self._do_read = injector.mem_read_double
        self._do_write = injector.mem_write_double
        super(DoubleArray, self).__init__(injector, 8, address, element_count, elements, auto_release)


class PointerArray(_ArrayType):

    def __init__(
            self,
            injector: object,
            address: int = 0,
            element_count: int = 0,
            elements: Optional[List[int]] = None,
            auto_release: bool = True
    ):
        address &= 0xFFFFFFFFFFFF
        self._do_read = injector.mem_read_pointer
        self._do_write = injector.mem_write_pointer
        elem_size = 4 if injector.bit_long == 32 else 8
        super(PointerArray, self).__init__(injector, elem_size, address, element_count, elements, auto_release)


class CString:
    """ This class's instance is returned by calling ``Injector.new_c_string`` """

    def __init__(self, injector: object, value: str, auto_release: bool = True):
        self._injector = injector
        self._value = value
        self._auto_release = auto_release

        self._address = self._injector.mem_alloc(alloc_size=len(value) + 1, protection="rw-")
        self._assign_value()

    def __del__(self):
        if self._auto_release:
            self.release()

    @property
    def value(self) -> str:
        """ string value """
        if self._address == 0:
            raise MemoryError(f"Memory has been freed")
        return self._injector.mem_read_c_string(self._address)

    @value.setter
    def value(self, value: str) -> None:
        raise NotImplementedError("CString is not editable")

    @property
    def address(self) -> int:
        """ string address """
        return self._address

    @property
    def auto_release(self) -> bool:
        """ auto release """
        return self._auto_release

    @auto_release.setter
    def auto_release(self, value: bool) -> None:
        self.set_auto_release(value)

    def set_auto_release(self, value: bool) -> None:
        """ Set whether to release automatically. """
        if not isinstance(value, bool):
            raise TypeError("auto_release should be bool")
        self._auto_release = value

    def _assign_value(self) -> None:
        if self._address == 0:
            raise MemoryError(f"Memory has been freed")
        self._injector.mem_write_c_string(self._address, self._value)

    def release(self) -> None:
        """ Release the memory. """
        if self._address:
            self._injector.mem_free(self._address)
            self._address = 0



from functools import wraps
from abc import ABCMeta, abstractmethod

from uniref.define.types import *
from uniref.define.patch import NativePatch


def _register_mem_read_array(proxy):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            address = args[1] if len(args) > 1 else kwargs.get("address")
            count = args[2] if len(args) > 2 else kwargs.get("count")
            if not isinstance(address, int):
                raise TypeError("address should be int")
            if count <= 0:
                raise ValueError("count should be positive")
            return proxy(args[0], address, count, auto_release=False).elements
        return inner
    return wrapper


def _register_mem_write_array(proxy):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            address = args[1] if len(args) > 1 else kwargs.get("address")
            elements = args[2] if len(args) > 2 else kwargs.get("elements")
            if not isinstance(address, int):
                raise TypeError("address should be int")
            if len(elements) <= 0:
                raise ValueError("Array is empty")
            proxy(args[0], address=address, elements=elements, auto_release=False)
            return True
        return inner
    return wrapper


def _register_new_basic_type(clazz: type, proxy):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            value = args[1] if len(args) > 1 else kwargs.get("value")
            auto_release = args[2] if len(args) > 2 else kwargs.get("auto_release", True)
            if not isinstance(value, clazz):
                raise TypeError(f"value should be {clazz.__name__}")
            if not isinstance(auto_release, bool):
                raise TypeError("auto_release should be bool")
            return proxy(*args, **kwargs)
        return inner
    return wrapper


def _register_new_array_type(clazz: type, proxy):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            value = args[1] if len(args) > 1 else kwargs.get("value")
            auto_release = args[2] if len(args) > 2 else kwargs.get("auto_release", True)
            if not all([isinstance(v, clazz) for v in value]):
                raise TypeError("value should all be int")
            if not isinstance(auto_release, bool):
                raise TypeError("auto_release should be bool")
            return proxy(args[0], elements=value, auto_release=auto_release)
        return inner
    return wrapper


class Injector(metaclass=ABCMeta):
    """
    Process injector interface.
    All read, write, apply, and release memory operations are completed in the injected process.
    """

    @property
    def bit_long(self) -> int:
        """ Application bit long """
        raise NotImplementedError

    @property
    def process_id(self) -> int:
        """ PID """
        raise NotImplementedError

    @property
    def code_compiler(self):
        raise NotImplementedError

    @abstractmethod
    def _mem_read(self, address: int, count: int) -> bytes:
        ...

    @abstractmethod
    def _mem_write(self, address: int, buffer: bytes) -> int:
        ...

    @abstractmethod
    def mem_alloc(self, alloc_size: int, protection: str = "rwx") -> int:
        """ Allocate memory.

        .. note::
            This method has different parameters on different platforms.
        """
        ...

    @abstractmethod
    def mem_free(self, address: int) -> None:
        """ Free allocated memory.

        .. note::
            This method has different parameters on different platforms.
        """
        ...

    @abstractmethod
    def mem_set_protect(self, address: int, length: int, new_protect: str) -> Optional[str]:
        """ Set memory protection.

        .. note::
            This method has different parameters on different platforms.
        """
        ...

    @abstractmethod
    def mem_read_bool(self, address: int) -> bool:
        """ Read a ``bool`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_char(self, address: int) -> int:
        """ Read a ``char`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_uchar(self, address: int) -> int:
        """ Read an ``unsigned char`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_int16(self, address: int) -> int:
        """ Read a ``short`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_uint16(self, address: int) -> int:
        """ Read an ``unsigned short`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_int32(self, address: int) -> int:
        """ Read an ``int`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_uint32(self, address: int) -> int:
        """ Read an ``unsigned int`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_int64(self, address: int) -> int:
        """ Read a ``long long`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_uint64(self, address: int) -> int:
        """ Read an ``unsigned long long`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_float(self, address: int) -> float:
        """ Read a ``float`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_double(self, address: int) -> float:
        """ Read a ``double`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_pointer(self, address: int) -> int:
        """ Read a ``void*`` value from the specified address. """
        ...

    @abstractmethod
    def mem_read_c_string(self, address: int) -> str:
        """ Read a C-style string from the specified address. """
        ...

    @abstractmethod
    def mem_write_bool(self, address: int, value: bool) -> bool:
        """ Write a ``bool`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_char(self, address: int, value: int) -> bool:
        """ Write a ``char`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_uchar(self, address: int, value: int) -> bool:
        """ Write an ``unsigned char`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_int16(self, address: int, value: int) -> bool:
        """ Write a ``short`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_uint16(self, address: int, value: int) -> bool:
        """ Write an ``unsigned short`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_int32(self, address: int, value: int) -> bool:
        """ Write an ``int`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_uint32(self, address: int, value: int) -> bool:
        """ Write an ``unsigned int`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_int64(self, address: int, value: int) -> bool:
        """ Write a ``long long`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_uint64(self, address: int, value: int) -> bool:
        """ Write an ``unsigned long long`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_float(self, address: int, value: float) -> bool:
        """ Write a ``float`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_double(self, address: int, value: float) -> bool:
        """ Write a ``double`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_pointer(self, address: int, value: int) -> int:
        """ Write a ``void*`` value to the specified address. """
        ...

    @abstractmethod
    def mem_write_c_string(self, address: int, value: str) -> bool:
        """ Write a C-style string to the specified address. (Automatically add ``\\x00``) """
        ...

    @abstractmethod
    def get_module_base(self, module_name: str) -> int:
        """ Get module base by name in the injected process. """
        ...

    @abstractmethod
    def get_proc_address(self) -> int:
        """ Get module exported function address in the injected process.

        .. note::
            This method has different parameters on different platforms.
        """
        ...

    @abstractmethod
    def call_native_function(self, func_address: int, args: Tuple[int], ret_type: int,
                             call_type: int, user_data: Optional[Any] = None) -> Any:
        ...

    @abstractmethod
    def code_nop(self, address: int, size: int) -> NativePatch:
        ...

    @_register_mem_read_array(proxy=BoolArray)
    def mem_read_bool_array(self, address: int, count: int) -> List[bool]:
        """ Read a ``bool`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=CharArray)
    def mem_read_char_array(self, address: int, count: int) -> List[int]:
        """ Read a ``char`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=UCharArray)
    def mem_read_uchar_array(self, address: int, count: int) -> List[int]:
        """ Read an ``unsigned char`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=Int16Array)
    def mem_read_int16_array(self, address: int, count: int) -> List[int]:
        """ Read a ``short`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=UInt16Array)
    def mem_read_uint16_array(self, address: int, count: int) -> List[int]:
        """ Read an ``unsigned short`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=Int32Array)
    def mem_read_int32_array(self, address: int, count: int) -> List[int]:
        """ Read an ``int`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=UInt32Array)
    def mem_read_uint32_array(self, address: int, count: int) -> List[int]:
        """ Read an ``unsigned int`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=Int64Array)
    def mem_read_int64_array(self, address: int, count: int) -> List[int]:
        """ Read a ``long long`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=UInt64Array)
    def mem_read_uint64_array(self, address: int, count: int) -> List[int]:
        """ Read an ``unsigned long long`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=FloatArray)
    def mem_read_float_array(self, address: int, count: int) -> List[float]:
        """ Read a ``float`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=DoubleArray)
    def mem_read_double_array(self, address: int, count: int) -> List[float]:
        """ Read a ``double`` array from the specified address. """
        ...

    @_register_mem_read_array(proxy=PointerArray)
    def mem_read_pointer_array(self, address: int, count: int) -> List[int]:
        """ Read a ``void*`` array from the specified address. """
        ...

    @_register_mem_write_array(proxy=BoolArray)
    def mem_write_bool_array(self, address: int, elements: List[int]) -> bool:
        """ Write a ``bool`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=CharArray)
    def mem_write_char_array(self, address: int, elements: List[int]) -> bool:
        """ Write a ``char`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=UCharArray)
    def mem_write_uchar_array(self, address: int, elements: List[int]) -> bool:
        """ Write an ``unsigned char`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=Int16Array)
    def mem_write_int16_array(self, address: int, elements: List[int]) -> bool:
        """ Write a ``short`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=UInt16Array)
    def mem_write_uint16_array(self, address: int, elements: List[int]) -> bool:
        """ Write an ``unsigned short`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=Int32Array)
    def mem_write_int32_array(self, address: int, elements: List[int]) -> bool:
        """ Write an ``int`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=UInt32Array)
    def mem_write_uint32_array(self, address: int, elements: List[int]) -> bool:
        """ Write an ``unsigned int`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=Int64Array)
    def mem_write_int64_array(self, address: int, elements: List[int]) -> bool:
        """ Write a ``long long`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=UInt64Array)
    def mem_write_uint64_array(self, address: int, elements: List[int]) -> bool:
        """ Write an ``unsigned long long`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=FloatArray)
    def mem_write_float_array(self, address: int, elements: List[float]) -> bool:
        """ Write a ``float`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=DoubleArray)
    def mem_write_double_array(self, address: int, elements: List[float]) -> bool:
        """ Write a ``double`` array to the specified address. """
        ...

    @_register_mem_write_array(proxy=PointerArray)
    def mem_write_pointer_array(self, address: int, elements: List[int]) -> bool:
        """ Write a ``void*`` array from to specified address. """
        ...

    @_register_new_basic_type(clazz=bool, proxy=Bool)
    def new_bool(self, value: bool, auto_release: bool = True) -> Bool:
        """ New a ``bool``

        :param value: initial value
        :param auto_release: whether to delete automatically

        If ``auto_release`` is set to ``True``, the life cycle of this newly allocated memory is the same as the ``Bool`` instance.
        Otherwise, you can release it manually later by calling ``Bool.release()``.

        Examples:

        .. code-block:: python

            from uniref import *

            ref = WinUniRef("game.exe")

            def auto_release_sample():
                bool_1 = ref.injector.new_bool(True)
                bool_2 = ref.injector.new_bool(True, auto_release=False)
                bool_3 = ref.injector.new_bool(True, auto_release=False)

                # ...

                bool_2.auto_release = True
                return bool_3

            manual_bool = auto_release_sample()
            # bool_1 and bool_2 have been released
            # now release bool_3 manually
            manual_bool.release()
        """
        ...

    @_register_new_basic_type(clazz=int, proxy=Char)
    def new_char(self, value: int, auto_release: bool = True) -> Char:
        """ New a ``char`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=int, proxy=UChar)
    def new_uchar(self, value: int, auto_release: bool = True) -> UChar:
        """ New an ``unsigned char`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=int, proxy=Int16)
    def new_int16(self, value: int, auto_release: bool = True) -> Int16:
        """ New a ``short`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=int, proxy=UInt16)
    def new_uint16(self, value: int, auto_release: bool = True) -> UInt16:
        """ New an ``unsigned short`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=int, proxy=Int32)
    def new_int32(self, value: int, auto_release: bool = True) -> Int32:
        """ New an ``int`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=int, proxy=UInt32)
    def new_uint32(self, value: int, auto_release: bool = True) -> UInt32:
        """ New an ``unsigned int`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=int, proxy=Int64)
    def new_int64(self, value: int, auto_release: bool = True) -> Int64:
        """ New a ``long long`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=int, proxy=UInt64)
    def new_uint64(self, value: int, auto_release: bool = True) -> UInt64:
        """ New an ``unsigned long long`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=float, proxy=Float)
    def new_float(self, value: float, auto_release: bool = True) -> Float:
        """ New a ``float`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=float, proxy=Double)
    def new_double(self, value: float, auto_release: bool = True) -> Double:
        """ New a ``double`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=int, proxy=Pointer)
    def new_pointer(self, value: int, auto_release: bool = True) -> Pointer:
        """ New a ``void*`` variable. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_basic_type(clazz=str, proxy=CString)
    def new_c_string(self, value: str, auto_release: bool = True) -> CString:
        """ New a C-style string. Check ``new_bool``'s description for more details. """
        ...

    @_register_new_array_type(clazz=bool, proxy=BoolArray)
    def new_bool_array(self, value: List[bool], auto_release: bool = True) -> BoolArray:
        """ New a ``bool`` array.

        :param value: list of initial values
        :param auto_release: whether to delete automatically

        If ``auto_release`` is set to ``True``, the life cycle of this newly allocated memory is the same as the ``BoolArray`` instance.
        Otherwise, you can release it manually later by calling ``BoolArray.release()``.

        Examples:

        .. code-block:: python

            from uniref import *

            ref = WinUniRef("game.exe")

            def auto_release_sample():
                array_1 = ref.injector.new_bool_array([True, False])
                array_2 = ref.injector.new_bool_array([True, False, True], auto_release=False)
                array_3 = ref.injector.new_bool_array([True, False, True, True], auto_release=False)

                # ...

                array_2.auto_release = True
                return array_3

            manual_array = auto_release_sample()
            # array_1 and array_2 have been released
            # now release array_3 manually
            manual_array.release()
        """
        ...

    @_register_new_array_type(clazz=int, proxy=CharArray)
    def new_char_array(self, value: List[int], auto_release: bool = True) -> CharArray:
        """ New a ``char`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=int, proxy=UCharArray)
    def new_uchar_array(self, value: List[int], auto_release: bool = True) -> UCharArray:
        """ New an ``unsigned char`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=int, proxy=Int16Array)
    def new_int16_array(self, value: List[int], auto_release: bool = True) -> Int16Array:
        """ New a ``short`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=int, proxy=UInt16Array)
    def new_uint16_array(self, value: List[int], auto_release: bool = True) -> UInt16Array:
        """ New an ``unsigned short`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=int, proxy=Int32Array)
    def new_int32_array(self, value: List[int], auto_release: bool = True) -> Int32Array:
        """ New an ``int`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=int, proxy=UInt32Array)
    def new_uint32_array(self, value: List[int], auto_release: bool = True) -> UInt32Array:
        """ New an ``unsigned int`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=int, proxy=Int64Array)
    def new_int64_array(self, value: List[int], auto_release: bool = True) -> Int64Array:
        """ New a ``long long`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=int, proxy=UInt64Array)
    def new_uint64_array(self, value: List[int], auto_release: bool = True) -> UInt64Array:
        """ New an ``unsigned long long`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=float, proxy=FloatArray)
    def new_float_array(self, value: List[float], auto_release: bool = True) -> FloatArray:
        """ New a ``float`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=float, proxy=DoubleArray)
    def new_double_array(self, value: List[float], auto_release: bool = True) -> DoubleArray:
        """ New a ``double`` array. Check ``new_bool_array``'s description for more details. """
        ...

    @_register_new_array_type(clazz=int, proxy=PointerArray)
    def new_pointer_array(self, value: List[int], auto_release: bool = True) -> PointerArray:
        """ New a ``void*`` array. Check ``new_bool_array``'s description for more details. """
        ...

    def _read_cs_string(self, address: int) -> str:
        if self.bit_long == 32:
            raise NotImplementedError("32-bit read System.String")
        else:
            klass = self.mem_read_uint64(address)
            length = self.mem_read_uint32(klass + 0x10)
            return self.mem_read_bytes(klass + 0x14, length * 2).decode("utf-16")

    def _code_patch_internal(self, code: str or bytes, address: int) -> NativePatch:
        if isinstance(code, str):
            code = self.code_compile(code, address)

        origin_code = self._mem_read(address, len(code))
        return NativePatch(self, address, code, origin_code)

    def mem_read_multilevel_pointer(self, base: int, offsets: List[int]) -> int:
        """ Read multilevel pointer. """
        if not isinstance(base, int):
            raise TypeError("base should be int")
        if any([not isinstance(o, int) for o in offsets]):
            raise TypeError("offsets array should all be int")
        if not offsets:
            raise ValueError("offsets array is empty")
        p = self.mem_read_pointer(base + offsets[0])
        for i in range(1, len(offsets)):
            p = self.mem_read_pointer(p + offsets[i])
        return p

    def mem_read_bytes(self, address: int, count: int) -> bytes:
        """ Read ``count`` bytes from the specified address. """
        if not isinstance(address, int):
            raise TypeError("address should be int")
        if not isinstance(count, int):
            raise TypeError("count should be int")
        if count <= 0:
            raise ValueError("count should be positive")
        return self._mem_read(address, count)

    def mem_write_bytes(self, address: int, value: bytes) -> int:
        """ Write a byte array to the specified address. """
        if not isinstance(address, int):
            raise TypeError("address should be int")
        if not isinstance(value, bytes):
            raise TypeError("value should be bytes")
        if len(value) == 0:
            raise ValueError("value length should be positive")
        return self._mem_write(address, value)

    def code_compile(self, code: str, address: int = 0) -> bytes:
        return self.code_compiler.compile(code, address)

    def code_patch(self, code: str or bytes, address: int) -> NativePatch:
        if not isinstance(code, (str, bytes)):
            raise TypeError("code should be str or bytes")
        if not isinstance(address, int):
            raise TypeError("address should be int")
        return self._code_patch_internal(code, address)



import os
import struct
from pathlib import Path
from functools import wraps
from subprocess import Popen, PIPE

from uniref.define.patch import NativePatch
from uniref.define.struct import *
from uniref.define.types import *
from uniref.mono.assets import *
from uniref.util.winapi import WinApi
from uniref.util.compiler import X86Compiler
from uniref.util.injector.interface import Injector


def _register_mem_read(size: int, fmt: str):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            address = args[1] if len(args) > 1 else kwargs.get("address")
            if not isinstance(address, int):
                raise TypeError("address should be int")
            raw = args[0]._mem_read(address, size)
            return struct.unpack(fmt, raw)[0]
        return inner
    return wrapper


def _register_mem_write(clazz: type, fmt: str):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            address = args[1] if len(args) > 1 else kwargs.get("address")
            if not isinstance(address, int):
                raise TypeError("address should be int")
            value = args[2] if len(args) > 2 else kwargs.get("value")
            if not isinstance(value, clazz):
                raise TypeError(f"value should be {clazz.__name__}")
            raw = struct.pack(fmt, value)
            return args[0]._mem_write(address, raw) == len(raw)
        return inner
    return wrapper


class WinInjector(Injector):
    """ Process injector for ``Windows``. """

    def __init__(self, exe_filename: str = '', process_id: int = 0) -> None:
        if not process_id:
            if exe_filename:
                process_id = WinApi.GetPidByName(exe_filename)
                if not process_id:
                    raise Exception(f"Process with filename [{exe_filename}] not found")
                if len(process_id) != 1:
                    raise Exception(f"Too many processes with the same file name, use process id instead")
                process_id = process_id[0]
            else:
                process_id = WinApi.GetCurrentProcessId()
        h_process = WinApi.OpenProcess(PROCESS_ALL_ACCESS, False, process_id)

        self._h_process = h_process
        self._process_id = process_id
        self._bit_long = WinApi.ProcessBitLong(h_process)
        self._code_compiler = X86Compiler(bit_long=self._bit_long)

        self._get_proc_address = self._get_kernel32_proc_address("GetProcAddress")

    def __del__(self):
        if self._h_process > 0:
            WinApi.CloseHandle(self._h_process)
        self._h_process = 0

    @property
    def bit_long(self) -> int:
        return self._bit_long

    @property
    def process_id(self) -> int:
        return self._process_id

    @property
    def code_compiler(self):
        return self._code_compiler

    @property
    def process_handle(self) -> int:
        return self._h_process

    def _mem_read(self, address: int, count: int) -> bytes:
        if WinApi.VAReadableSize(self._h_process, address):
            n_read = c_int(0)
            buffer = (c_char * count)()
            WinApi.ReadProcessMemory(self._h_process, address, buffer, count, byref(n_read))
            return buffer.raw
        raise MemoryError(f"Address {hex(address)} not readable")

    def _mem_write(self, address: int, buffer: bytes) -> int:
        if WinApi.VAWriteableSize(self._h_process, address):
            n_write = c_int(0)
            WinApi.WriteProcessMemory(self._h_process, address, buffer, len(buffer), byref(n_write))
            return n_write.value
        raise MemoryError(f"Address {hex(address)} not writeable")

    def _get_kernel32_proc_address(self, proc_name: str):
        exe_dir = str(Path(os.path.abspath(__file__)).parent.parent.parent / f"bin/win/")
        p = Popen(f"getproc{self._bit_long}.exe {proc_name}", shell=True, cwd=exe_dir, stdout=PIPE)
        output = p.stdout.read()
        p.wait()
        if isinstance(output, bytes):
            output = output.decode(errors="ignore")
        proc_address = int(output, 16)
        if proc_address <= 0:
            raise SystemError(f"Can't get address of {proc_name}")
        return proc_address

    def _create_remote_thread(self, address: int) -> None:
        h_thread = WinApi.CreateRemoteThread(self._h_process, 0, 0, address, 0, 0, 0)
        WinApi.WaitForSingleObject(h_thread, INFINITE)
        WinApi.CloseHandle(h_thread)

    def mem_alloc(self,
                  alloc_address: int = 0,
                  alloc_size: int = PAGE_SIZE,
                  alloc_type: int = MEM_COMMIT | MEM_RESERVE,
                  protection: str = "rwx"
                  ) -> int:
        """ Allocate memory. """
        protection = protection_map.get(protection, PAGE_EXECUTE_READWRITE)
        return WinApi.VirtualAllocEx(self._h_process, alloc_address, alloc_size, alloc_type, protection)

    def mem_free(self, address: int, free_size: int = 0, free_type: int = MEM_RELEASE) -> None:
        """ Free allocated memory. """
        WinApi.VirtualFreeEx(self._h_process, address, free_size, free_type)

    def mem_set_protect(self, address: int, length: int, new_protect: str = "rwx") -> str:
        """ Set memory protection. """
        old = c_uint32()
        protection = protection_map.get(new_protect, PAGE_EXECUTE_READWRITE)
        WinApi.VirtualProtectEx(self._h_process, address, length, protection, byref(old))
        return protection_rev_map.get(old.value, "rwx")

    @_register_mem_read(size=1, fmt='?')
    def mem_read_bool(self, address: int) -> bool:
        ...

    @_register_mem_read(size=1, fmt='b')
    def mem_read_char(self, address: int) -> int:
        ...

    @_register_mem_read(size=1, fmt='B')
    def mem_read_uchar(self, address: int) -> int:
        ...

    @_register_mem_read(size=2, fmt='h')
    def mem_read_int16(self, address: int) -> int:
        ...

    @_register_mem_read(size=2, fmt='H')
    def mem_read_uint16(self, address: int) -> int:
        ...

    @_register_mem_read(size=4, fmt='i')
    def mem_read_int32(self, address: int) -> int:
        ...

    @_register_mem_read(size=4, fmt='I')
    def mem_read_uint32(self, address: int) -> int:
        ...

    @_register_mem_read(size=8, fmt='q')
    def mem_read_int64(self, address: int) -> int:
        ...

    @_register_mem_read(size=8, fmt='Q')
    def mem_read_uint64(self, address: int) -> int:
        ...

    @_register_mem_read(size=4, fmt='f')
    def mem_read_float(self, address: int) -> float:
        ...

    @_register_mem_read(size=8, fmt='d')
    def mem_read_double(self, address: int) -> float:
        ...

    def mem_read_pointer(self, address: int) -> int:
        if not isinstance(address, int):
            raise TypeError("address should be int")
        if self._bit_long == 32:
            return self.mem_read_uint32(address)
        else:
            return self.mem_read_uint64(address)

    def mem_read_c_string(self, address: int) -> str:
        if not isinstance(address, int):
            raise TypeError("address should be int")

        string = b''
        while True:
            char = self._mem_read(address + len(string), 1)
            if char == b'\x00':
                break
            string += char
        return string.decode()

    @_register_mem_write(clazz=bool, fmt='?')
    def mem_write_bool(self, address: int, value: bool) -> bool:
        ...

    @_register_mem_write(clazz=int, fmt='b')
    def mem_write_char(self, address: int, value: int) -> bool:
        ...

    @_register_mem_write(clazz=int, fmt='B')
    def mem_write_uchar(self, address: int, value: int) -> bool:
        ...

    @_register_mem_write(clazz=int, fmt='h')
    def mem_write_int16(self, address: int, value: int) -> bool:
        ...

    @_register_mem_write(clazz=int, fmt='H')
    def mem_write_uint16(self, address: int, value: int) -> bool:
        ...

    @_register_mem_write(clazz=int, fmt='i')
    def mem_write_int32(self, address: int, value: int) -> bool:
        ...

    @_register_mem_write(clazz=int, fmt='I')
    def mem_write_uint32(self, address: int, value: int) -> bool:
        ...

    @_register_mem_write(clazz=int, fmt='q')
    def mem_write_int64(self, address: int, value: int) -> bool:
        ...

    @_register_mem_write(clazz=int, fmt='Q')
    def mem_write_uint64(self, address: int, value: int) -> bool:
        ...

    @_register_mem_write(clazz=float, fmt='f')
    def mem_write_float(self, address: int, value: float) -> bool:
        ...

    @_register_mem_write(clazz=float, fmt='d')
    def mem_write_double(self, address: int, value: float) -> bool:
        ...

    def mem_write_pointer(self, address: int, value: int) -> int:
        if not isinstance(address, int):
            raise TypeError("address should be int")
        if not isinstance(value, int):
            raise TypeError("value should be int")
        if self._bit_long == 32:
            return self.mem_write_uint32(address, value)
        else:
            return self.mem_write_uint64(address, value)

    def mem_write_c_string(self, address: int, value: str) -> bool:
        if not isinstance(address, int):
            raise TypeError("address should be int")
        if not isinstance(value, str):
            raise TypeError("value should be str")
        if len(value) == 0:
            to_write = b'\x00'
        else:
            to_write = value.encode(errors="ignore") + b'\x00'
        return self.mem_write_bytes(address, to_write) == len(value) + 1

    def code_execute(self, code: str or bytes, target_address: int = 0) -> int:
        if not isinstance(code, (str, bytes)):
            raise TypeError("code should be str or bytes")
        if not isinstance(target_address, int):
            raise TypeError("target_address should be int")
        if isinstance(code, str):
            code = self.code_compile(code, target_address)

        address = target_address if target_address else self.mem_alloc(alloc_size=len(code))
        self._code_patch_internal(code, address)
        self._create_remote_thread(address)
        return address

    def get_module_base(self, module_name: str) -> int:
        if not isinstance(module_name, str):
            raise TypeError("module_name should be str")
        return WinApi.GetRemoteModuleHandle(self._h_process, module_name)

    def get_proc_address(self, module_base: int, proc_name: str) -> int:
        """ Get module exported function address in the injected process. """
        if not isinstance(module_base, int):
            raise TypeError("module_base should be int")
        if not isinstance(proc_name, str):
            raise TypeError("proc_name should be str")
        if module_base <= 0:
            raise ValueError("module_base should be positive")
        if len(proc_name) > 255:
            raise ValueError("proc_name too long")

        page_start = self.mem_alloc()
        return_address = page_start + 0x300
        proc_name_address = page_start + 0x200
        if self._bit_long == 32:
            code = f"push {hex(proc_name_address)}           \n" \
                   f"push {hex(module_base)}                 \n" \
                   f"mov eax, {hex(self._get_proc_address)}  \n" \
                   f"call eax                                \n" \
                   f"mov ecx, {hex(return_address)}          \n" \
                   f"mov dword ptr [ecx], eax                \n" \
                   f"ret"
        else:
            code = f"sub rsp, 28h                            \n" \
                   f"mov rcx, {hex(module_base)}             \n" \
                   f"mov rdx, {hex(proc_name_address)}       \n" \
                   f"mov rax, {hex(self._get_proc_address)}  \n" \
                   f"call rax                                \n" \
                   f"mov r12, {hex(return_address)}          \n" \
                   f"mov qword ptr [r12], rax                \n" \
                   f"add rsp, 28h                            \n" \
                   f"ret"

        self._mem_write(proc_name_address, proc_name.encode(errors="ignore") + b'\x00')
        self.code_execute(code, page_start)
        proc_address = self.mem_read_pointer(return_address)
        self.mem_free(page_start)
        return proc_address

    def call_native_function(self, func_address: int, args: Tuple[int], ret_type: int,
                             call_type: int, user_data: Optional[Any] = None) -> Any:
        if not isinstance(func_address, int):
            raise TypeError("func_address should be int")
        if not isinstance(args, tuple):
            raise TypeError("args should be a tuple of int")
        if not isinstance(ret_type, int):
            raise TypeError("ret_type should be int")
        if not isinstance(call_type, int):
            raise TypeError("call_type should be int")
        if func_address <= 0:
            raise ValueError("func_address should be positive")
        if ret_type not in range(TYPE_CS_STRING + 1):
            raise ValueError(f"Unsupported return type")
        if call_type not in range(CALL_TYPE_FASTCALL + 1):
            raise ValueError(f"Unsupported call convention")

        page_start = self.mem_alloc()
        return_address = page_start + 0x300

        if self._bit_long == 32:
            code = "push ebp\n mov ebp, esp\n"

            if isinstance(user_data, list) and all([i != 0 for i in user_data]):
                domain, attach, detach = user_data
                thread_address = page_start + 0x400

                code += f"push {hex(domain)}              \n" \
                        f"mov eax, {hex(attach)}          \n" \
                        f"call eax                        \n" \
                        f"mov ecx, {hex(thread_address)}  \n" \
                        f"mov dword ptr [ecx], eax        \n" \

            if call_type == CALL_TYPE_THISCALL:
                if len(args) < 1:
                    raise ValueError("thiscall needs [this] pointer")
                code += f"mov ecx, {hex(args[0])}      \n"
                args = args[1:]

            if call_type == CALL_TYPE_FASTCALL:
                argc = len(args)
                if argc:
                    code += f"mov ecx, {hex(args[0])}"
                if argc >= 2:
                    code += f"mov edx, {hex(args[1])}"
                args = args[2:]

            for arg in args[::-1]:
                code += f"push {hex(arg)}              \n"

            code += f"mov eax, {hex(func_address)}     \n" \
                    f"call eax                         \n" \
                    f"mov ecx, {hex(return_address)}   \n" \
                    f"mov dword ptr [ecx], eax         \n" \

            if isinstance(user_data, list) and all([i != 0 for i in user_data]):
                code += f"mov ecx, {hex(thread_address)}  \n" \
                        f"push dword ptr [ecx]            \n" \
                        f"mov eax, {hex(detach)}          \n" \
                        f"call eax                        \n" \

            code += f"leave\n ret"
        else:
            frame_size = 0x28 + ((len(args[4:]) + 1) // 2) * 0x10
            code = f"sub rsp, {hex(frame_size)}        \n"

            if isinstance(user_data, list) and all([i != 0 for i in user_data]):
                domain, attach, detach = user_data
                thread_address = page_start + 0x400

                code += f"mov rcx, {hex(domain)}          \n" \
                        f"mov rax, {hex(attach)}          \n" \
                        f"call rax                        \n" \
                        f"mov r12, {hex(thread_address)}  \n" \
                        f"mov qword ptr [r12], rax        \n" \

            regs = ["rcx", "rdx", "r8", "r9"]
            for idx, arg in enumerate(args[:4]):
                code += f"mov {regs[idx]}, {hex(arg)}  \n"

            args = args[4:]
            for i in range(len(args)):
                code += f"mov qword ptr [rsp+{hex(0x20+i*8)}], {hex(args[i])}  \n"

            code += f"mov rax, {hex(func_address)}     \n" \
                    f"call rax                         \n" \
                    f"mov r12, {hex(return_address)}   \n" \
                    f"mov qword ptr [r12], rax         \n" \

            if isinstance(user_data, list) and all([i != 0 for i in user_data]):
                code += f"mov r12, {hex(thread_address)}  \n" \
                        f"mov rcx, qword ptr [r12]        \n" \
                        f"mov rax, {hex(detach)}          \n" \
                        f"call rax                        \n" \

            code += f"add rsp, {hex(frame_size)}       \n" \
                    f"ret"

        code = self.code_compile(code)
        n_write = self.mem_write_bytes(page_start, code)
        if n_write != len(code):
            raise OSError("write compiled code to memory", f"{n_write}/{len(code)} bytes written")
        self._create_remote_thread(page_start)

        if ret_type == TYPE_BOOL:
            ret = self.mem_read_bool(return_address)
        elif ret_type == TYPE_CHAR:
            ret = self.mem_read_char(return_address)
        elif ret_type == TYPE_UCHAR:
            ret = self.mem_read_uchar(return_address)
        elif ret_type == TYPE_INT16:
            ret = self.mem_read_int16(return_address)
        elif ret_type == TYPE_UINT16:
            ret = self.mem_read_uint16(return_address)
        elif ret_type == TYPE_INT32:
            ret = self.mem_read_int32(return_address)
        elif ret_type == TYPE_UINT32:
            ret = self.mem_read_uint32(return_address)
        elif ret_type == TYPE_INT64:
            ret = self.mem_read_int64(return_address)
        elif ret_type == TYPE_UINT64:
            ret = self.mem_read_uint64(return_address)
        elif ret_type == TYPE_FLOAT:
            ret = self.mem_read_float(return_address)
        elif ret_type == TYPE_DOUBLE:
            ret = self.mem_read_double(return_address)
        elif ret_type == TYPE_VOID_P:
            ret = self.mem_read_pointer(return_address)
        elif ret_type == TYPE_CHAR_P:
            char_address = self.mem_read_pointer(return_address)
            ret = self.mem_read_c_string(char_address)
        elif ret_type == TYPE_CS_STRING:
            ret = self._read_cs_string(return_address)
        else:
            ret = None
        self.mem_free(page_start)
        return ret

    def code_nop(self, address: int, size: int) -> NativePatch:
        return self.code_patch(b'\x90' * size, address)



import win32process
from typing import *
from functools import wraps
from platform import architecture

from ctypes import windll
from uniref.define.struct import *
from uniref.define.constant import *


def _register_api(unexpected_return: object):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            api_name = func.__name__
            ret = eval(f"windll.kernel32.{api_name}(*args, **kwargs)")
            if ret == unexpected_return:
                raise SystemError(f"[{api_name}] error code: {WinApi.GetLastError()}")
            return ret
        return inner
    return wrapper


def _register_api_without_check(func):
    @wraps(func)
    def inner(*args, **kwargs):
        api_name = func.__name__
        return eval(f"windll.kernel32.{api_name}(*args, **kwargs)")
    return inner


class WinApi:
    _kernel32 = windll.kernel32
    _kernel32.CloseHandle.restype = c_bool
    _kernel32.Process32First.restype = c_bool
    _kernel32.Process32Next.restype = c_bool
    _kernel32.Module32First.restype = c_bool
    _kernel32.Module32Next.restype = c_bool
    _kernel32.ReadProcessMemory.restype = c_bool
    _kernel32.ReadProcessMemory.argtypes = [c_void_p, c_void_p, c_void_p, c_size_t, c_void_p]
    _kernel32.WriteProcessMemory.restype = c_bool
    _kernel32.WriteProcessMemory.argtypes = [c_void_p, c_void_p, c_void_p, c_size_t, c_void_p]
    _kernel32.GetProcAddress.restype = c_void_p
    _kernel32.GetProcAddress.argtypes = [c_void_p, c_char_p]
    _kernel32.GetModuleHandleA.restype = c_void_p
    _kernel32.CreateRemoteThread.argtypes = [c_void_p, c_void_p, c_size_t, c_void_p, c_void_p, c_uint32, c_void_p]
    _kernel32.VirtualQueryEx.restype = c_uint64
    _kernel32.VirtualQueryEx.argtypes = [c_void_p, c_void_p, c_void_p, c_uint32]
    _kernel32.VirtualProtectEx.restype = c_bool
    _kernel32.VirtualProtectEx.argtypes = [c_void_p, c_void_p, c_size_t, c_uint32, c_void_p]
    _kernel32.VirtualAllocEx.restype = c_void_p
    _kernel32.VirtualAllocEx.argtypes = [c_void_p, c_void_p, c_uint64, c_uint32, c_uint32]
    _kernel32.VirtualFreeEx.restype = c_bool
    _kernel32.VirtualFreeEx.argtypes = [c_void_p, c_void_p, c_uint64, c_uint32]

    _is_x86_32 = "32" in architecture()[0]
    _readable_blacklist = (PAGE_NOACCESS, PAGE_EXECUTE, PAGE_NOCACHE, PAGE_GUARD)
    _writeable_whitelist = (PAGE_READWRITE, PAGE_WRITECOPY, PAGE_EXECUTE_READWRITE, PAGE_EXECUTE_WRITECOPY, PAGE_WRITECOMBINE)

    @staticmethod
    def GetLastError() -> int:
        return windll.kernel32.GetLastError()

    @staticmethod
    @_register_api(unexpected_return=0)
    def GetCurrentProcess() -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=0)
    def GetCurrentProcessId() -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=None)
    def GetModuleHandleA(lpModuleName: bytes) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=0)
    def IsWow64Process(hProcess: int, Wow64Process: object) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=0)
    def TlsSetValue(dwTlsIndex: int, lpTlsValue: int) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=0)
    def OpenProcess(dwDesiredAccess: int, bInheritHandle: bool, dwProcessId: int) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=0)
    def VirtualQueryEx(hProcess: int, lpAddress: int, lpBuffer: object, dwLength: int) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=False)
    def VirtualProtectEx(hProcess: int, lpAddress: int, dwSize: int, flNewProtect: int, lpflOldProtect: object) -> bool:
        ...

    @staticmethod
    @_register_api(unexpected_return=None)
    def VirtualAllocEx(hProcess: int, lpAddress: int, dwSize: int, flAllocationType: int, flProtect: int) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=False)
    def VirtualFreeEx(hProcess: int, lpAddress: int, dwSize: int, dwFreeType: int) -> bool:
        ...

    @staticmethod
    @_register_api(unexpected_return=0)
    def CreateRemoteThread(hProcess: int, lpThreadAttributes: int, dwStackSize: int, lpStartAddress: int, lpParameter: int, dwCreationFlags: int, lpThreadId: object) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=-1)
    def WaitForSingleObject(hHandle: int, dwMilliseconds: int) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=None)
    def GetProcAddress(hModule: int, lpProcName: bytes) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=INVALID_HANDLE_VALUE)
    def CreateToolhelp32Snapshot(dwFlags: int, th32ProcessID: int) -> int:
        ...

    @staticmethod
    @_register_api(unexpected_return=False)
    def CloseHandle(hObject: int) -> bool:
        ...

    @staticmethod
    @_register_api_without_check
    def Process32First(hSnapshot: int, lppe: object) -> bool:
        ...

    @staticmethod
    @_register_api_without_check
    def Process32Next(hSnapshot: int, lppe: object) -> bool:
        ...

    @staticmethod
    @_register_api_without_check
    def Module32First(hSnapshot: int, lpme: object) -> bool:
        ...

    @staticmethod
    @_register_api_without_check
    def Module32Next(hSnapshot: int, lpme: object) -> bool:
        ...

    @staticmethod
    @_register_api(unexpected_return=False)
    def ReadProcessMemory(hProcess: int, lpBaseAddress: int, lpBuffer: object, nSize: int, lpNumberOfBytesRead: object) -> bool:
        ...

    @staticmethod
    @_register_api(unexpected_return=False)
    def WriteProcessMemory(hProcess: int, lpBaseAddress: int, lpBuffer: bytes, nSize: int, lpNumberOfBytesWritten: object) -> bool:
        ...

    @staticmethod
    def GetPidByName(process_name: str) -> List[int]:
        pid = []
        process_name = process_name.encode()

        pe = PROCESSENTRY32()
        pe.dwSize = sizeof(pe)
        h_process_snapshot = WinApi.CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0)
        ret = WinApi.Process32First(h_process_snapshot, byref(pe))
        while ret:
            if process_name == pe.szExeFile:
                pid.append(pe.th32ProcessID)
            ret = WinApi.Process32Next(h_process_snapshot, byref(pe))
        WinApi.CloseHandle(h_process_snapshot)
        return pid

    @staticmethod
    def GetRemoteModules(h_process: int) -> List[int]:
        bit_long = WinApi.ProcessBitLong(h_process)
        flag = LIST_MODULES_32BIT if bit_long == 32 else LIST_MODULES_64BIT
        return win32process.EnumProcessModulesEx(h_process, flag)

    @staticmethod
    def GetRemoteModuleHandle(h_process: int, module_name: str) -> int:
        modules = WinApi.GetRemoteModules(h_process)
        for module in modules:
            file_name = win32process.GetModuleFileNameEx(h_process, module)
            if file_name and file_name.endswith(module_name):
                return module
        return 0

    @staticmethod
    def ProcessBitLong(hProcess: int) -> int:
        if WinApi._is_x86_32:
            return 32

        flag = c_bool()
        WinApi.IsWow64Process(hProcess, byref(flag))
        return 32 if flag.value else 64

    @staticmethod
    def VAReadableSize(hProcess: int, address: int) -> int:
        if WinApi._is_x86_32:
            mbi = MEMORY_BASIC_INFORMATION32()
        else:
            mbi = MEMORY_BASIC_INFORMATION64()
        WinApi.VirtualQueryEx(hProcess, address, byref(mbi), sizeof(mbi))
        protect = mbi.Protect
        if protect:
            if not any([protect & flag for flag in WinApi._readable_blacklist]):
                return mbi.RegionSize
        return 0

    @staticmethod
    def VAWriteableSize(hProcess: int, address: int) -> int:
        if WinApi._is_x86_32:
            mbi = MEMORY_BASIC_INFORMATION32()
        else:
            mbi = MEMORY_BASIC_INFORMATION64()
        WinApi.VirtualQueryEx(hProcess, address, byref(mbi), sizeof(mbi))
        protect = mbi.Protect
        if protect:
            if any([protect & flag for flag in WinApi._writeable_whitelist]):
                return mbi.RegionSize
        return 0



from ctypes import *


class PROCESSENTRY32(Structure):
    _fields_ = [
        ("dwSize", c_ulong),
        ("cntUsage", c_ulong),
        ("th32ProcessID", c_ulong),
        ("th32DefaultHeapID", c_void_p),
        ("th32ModuleID", c_ulong),
        ("cntThreads", c_ulong),
        ("th32ParentProcessID", c_ulong),
        ("pcPriClassBase", c_ulong),
        ("dwFlags", c_ulong),
        ("szExeFile", c_char * 260)
    ]


class MODULEENTRY32(Structure):
    _fields_ = [
        ("dwSize", c_ulong),
        ("th32ModuleID", c_ulong),
        ("th32ProcessID", c_ulong),
        ("GlblcntUsage", c_ulong),
        ("ProccntUsage", c_ulong),
        ("modBaseAddr", c_char_p),
        ("modBaseSize", c_ulong),
        ("hModule", c_void_p),
        ("szModule", c_char * 256),
        ("szExePath", c_char * 260),
    ]


class MEMORY_BASIC_INFORMATION32(Structure):
    _fields_ = [
        ("BaseAddress", c_uint32),
        ("AllocationBase", c_uint32),
        ("AllocationProtect", c_uint32),
        ("RegionSize", c_uint32),
        ("State", c_uint32),
        ("Protect", c_uint32),
        ("Type", c_uint32),
    ]


class MEMORY_BASIC_INFORMATION64(Structure):
    _fields_ = [
        ("BaseAddress", c_ulonglong),
        ("AllocationBase", c_ulonglong),
        ("AllocationProtect", c_ulong),
        ("__alignment1", c_ulong),
        ("RegionSize", c_ulonglong),
        ("State", c_ulong),
        ("Protect", c_ulong),
        ("Type", c_ulong),
        ("__alignment2", c_ulong),
    ]


class CUSTOM_DOMAIN_ARRAY32(Structure):
    _fields_ = [
        ("cnt", c_uint32),
        ("domains", c_uint32 * 1023)
    ]


class CUSTOM_DOMAIN_ARRAY64(Structure):
    _fields_ = [
        ("cnt", c_uint32),
        ("domains", c_uint64 * 511)
    ]



""" This module holds the uniref top-level classes for all supported platforms """
import frida.core

from uniref.mono.component import *
from uniref.mono.injector import WinMonoInjector, AndroidMonoInjector


class _UniRef:

    @property
    def injector(self) -> WinMonoInjector:
        """ target process injector """
        return self._mono_injector

    @property
    def pid(self) -> int:
        """ target process id """
        return self.injector.process_id

    @property
    def mono_module_handle(self) -> int:
        """ target process mono / il2cpp module handle """
        return self.injector.h_mono

    @property
    def root_domain(self) -> int:
        """ target process root application domain """
        return self.injector.root_domain

    def use_il2cpp(self) -> bool:
        """ Check if target process uses the ``IL2CPP`` scripting backend.

        :return: ``False`` if target process uses the ``Mono`` scripting backend, else ``True``.
        """
        return self.injector.use_il2cpp

    def list_assemblies(self) -> List[MonoAssembly]:
        """ List all assemblies in the application. """
        return self.injector.enum_assemblies()

    def get_assembly_image(self, assembly: MonoAssembly) -> MonoImage:
        """ Get the image of the specified assembly. """
        if not isinstance(assembly, MonoAssembly):
            raise TypeError("assembly should be MonoAssembly")
        return self.injector.get_assembly_image(assembly.handle)

    def list_images(self) -> List[MonoImage]:
        """ List all images in the application. """
        return self.injector.enum_images()

    def find_image_by_name(self, image_name: str) -> Optional[MonoImage]:
        """ Find the image by image name.

        :return: ``MonoImage`` instance if image is found, else ``None``.
        """
        if not isinstance(image_name, str):
            raise TypeError("image_name should be str")
        return self.injector.find_image_by_name(image_name)

    def list_classes_in_image(self, image: MonoImage) -> List[MonoClass]:
        """ List all classes in the image. """
        if not isinstance(image, MonoImage):
            raise TypeError("image should be MonoImage")
        return self.injector.enum_classes_in_image(image.handle)

    def find_class_in_image(self, image_name: str, class_path: str) -> Optional[MonoClass]:
        """ Find the class in the image.

        Probably the most commonly used method.
        This method looks up class reflection information through
        the given image name and class full path ([namespace].[class name]).

        This method supports resolving the classpath shown in Cheat Engine.
        For **inner** classes, you can concatenate the class names by ``+`` or ``/``.

        Examples:

        >>> ref.find_class_in_image("Assembly-CSharp", "PlayerStats")
        <uniref.mono.component.MonoClass object at 0x0000022F53050CF8>
        >>> ref.find_class_in_image("Assembly-CSharp", "Rewired.UI.ControlMapper.Window+Timer")
        <uniref.mono.component.MonoClass object at 0x0000022F53050C50>
        >>> ref.find_class_in_image("Assembly-CSharp", "MyClass") is None
        True

        :return: ``MonoClass`` instance if class is found, else ``None``.
        """
        if not isinstance(image_name, str):
            raise TypeError("image_name should be str")
        if not isinstance(class_path, str):
            raise TypeError("class_path should be str")
        image = self.find_image_by_name(image_name)
        if image:
            if '.' in class_path:
                dot = class_path.rfind('.')
                class_namespace = class_path[:dot]
                class_name = class_path[dot + 1:]
            else:
                class_namespace = ''
                class_name = class_path
            class_name = class_name.replace('+', '/')
            return self.injector.find_class_in_image(image.handle, class_namespace, class_name)
        return None

    def list_fields_in_class(self, klass: MonoClass) -> List[MonoField]:
        """ List all fields in the class. """
        if not isinstance(klass, MonoClass):
            raise TypeError("klass should be MonoClass")
        fields = self.injector.enum_fields_in_class(klass.handle)
        instance = klass.instance
        if instance > 0:
            for field in fields:
                field.set_instance(instance)
        return fields

    def find_field_in_class(self, klass: MonoClass, field_name: str) -> Optional[MonoField]:
        """ Find the field by its name.

        :return: ``MonoField`` instance if field is found, else ``None``.
        """
        if not isinstance(klass, MonoClass):
            raise TypeError("klass should be MonoClass")
        if not isinstance(field_name, str):
            raise TypeError("field_name should be str")
        field = self.injector.find_field_in_class(klass.handle, field_name)
        if field:
            if klass.instance > 0:
                field.set_instance(klass.instance)
            return field

    def list_methods_in_class(self, klass: MonoClass) -> List[MonoMethod]:
        """ List all methods in the class. """
        if not isinstance(klass, MonoClass):
            raise TypeError("klass should be MonoClass")
        methods = self.injector.enum_methods_in_class(klass.handle)
        instance = klass.instance
        if instance > 0:
            for method in methods:
                method.set_instance(instance)
        return methods

    def find_method_in_class(self, klass: MonoClass, method_name: str, param_count: int = -1) -> Optional[MonoMethod]:
        """ Find the method in the class.

        If there are different overloads of a method in the same class,
        you can distinguish them by the number of parameters. This is what the ``param_count`` means.

        :return: ``MonoMethod`` instance if method is found, else ``None``.
        """
        if not isinstance(klass, MonoClass):
            raise TypeError("klass should be MonoClass")
        if not isinstance(method_name, str):
            raise TypeError("method_name should be str")
        if not isinstance(param_count, int):
            raise TypeError("param_count should be int")
        method = self.injector.find_method_in_class(klass.handle, method_name, param_count)
        if method:
            if klass.instance > 0:
                method.set_instance(klass.instance)
            return method


class WinUniRef(_UniRef):
    """ The uniref top-level class for ``Windows``.

    :param exe_filename: target application exe filename
    :param process_id: process id

    Examples:

    >>> ref = WinUniRef("TheForest.exe")
    >>> ref = WinUniRef(process_id=1234)

    """
    def __init__(self, exe_filename: str = '', process_id: int = 0) -> None:
        if not exe_filename and not process_id:
            raise ValueError("Please specify the exe file name or process id")
        self._mono_injector = WinMonoInjector(exe_filename, process_id)

    @property
    def process_handle(self) -> int:
        """ target process handle """
        return self.injector.process_handle


class AndroidUniRef(_UniRef):
    """ The uniref top-level class for ``Android``.

    :param process_name: the process name of the target application, you can get it by ``frida-ps``
    :param package_name: the package name of the target application
    :param device_id: specify the device id, you can get it by ``adb devices``
    :param spawn: decide whether to create or attach a process when ``package_name`` is given, ``True`` by default

    Examples:

    .. code-block:: python

        # automatically attach the frontmost application & device
        ref = AndroidUniRef()

        # attach the application by its process name (automatically select device)
        ref = AndroidUniRef(process_name="My App")

        # spawn the application by its package name (automatically select device)
        ref = AndroidUniRef(package_name="com.test.my_app")

        # attach the application by its package name (automatically select device)
        ref = AndroidUniRef(package_name="com.test.my_app", spawn=False)

        # attach the application on the specified device
        ref = AndroidUniRef(device_id="12a34b5")
        ref = AndroidUniRef(process_name="My App", device_id="12a34b5")
        ref = AndroidUniRef(package_name="com.test.my_app", device_id="12a34b5")

    """
    def __init__(
        self,
        process_name: Optional[str] = None,
        package_name: Optional[str] = None,
        device_id: Optional[str] = None,
        spawn: bool = True
    ) -> None:
        self._mono_injector = AndroidMonoInjector(process_name, package_name, device_id, spawn)

    def execute_js(self, js_code: str, on_message_callback: Optional[Callable] = None) -> frida.core.Script:
        """ inject javascript through ``Frida``.

        :param js_code: javascript code
        :param on_message_callback: callback in the form of **on_message(message, data)**
        :return: ``frida.core.Script`` instance
        """
        return self._mono_injector.execute_js(js_code, on_message_callback)



import os
import time
import frida
import struct
from typing import *
from pathlib import Path
from functools import wraps

from uniref.define.constant import *
from uniref.define.patch import NativePatch
from uniref.mono.assets import frida_type_map
from uniref.util.compiler import ArmCompiler
from uniref.util.injector.interface import Injector


def _register_mem_read(func):
    @wraps(func)
    def inner(*args, **kwargs):
        address = args[1] if len(args) > 1 else kwargs.get("address")
        if not isinstance(address, int):
            raise TypeError("address should be int")
        return func(*args, **kwargs)
    return inner


def _register_mem_write(clazz: type):
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            address = args[1] if len(args) > 1 else kwargs.get("address")
            if not isinstance(address, int):
                raise TypeError("address should be int")
            value = args[2] if len(args) > 2 else kwargs.get("value")
            if not isinstance(value, clazz):
                raise TypeError(f"value should be {clazz.__name__}")
            return isinstance(func(*args, **kwargs), str)
        return inner
    return wrapper


class AndroidInjector(Injector):

    def __init__(self,  process_name: Optional[str], package_name: Optional[str], device_id: Optional[str], spawn: bool) -> None:
        self._agent = None
        self._session = None
        self._bit_long = 0
        self._process_id = 0

        self._device = self._find_device(device_id)
        self._session = self._attach_application(process_name, package_name, spawn)

        script_path = Path(os.path.abspath(__file__)).parent.parent.parent / "bin/android/agent.js"
        self._agent = self._session.create_script(script_path.read_text("utf-8", "ignore"))
        self._agent.load()
        self._api = self._agent.exports

        self._code_compiler = ArmCompiler(bit_long=self.bit_long)

    @property
    def bit_long(self) -> int:
        if not self._bit_long:
            self._bit_long = self._api.pointer_size() * 8
        return self._bit_long

    @property
    def process_id(self) -> int:
        if not self._process_id:
            self._process_id = self._api.pid()
        return self._process_id

    @property
    def code_compiler(self):
        return self._code_compiler

    def _find_device(self, device_id: Optional[str]) -> frida.core.Device:
        if device_id:
            device = frida.get_device(device_id)
        else:
            device = frida.get_usb_device()
        return device

    def _attach_application(self, process_name: Optional[str], package_name: Optional[str], spawn: bool) -> Optional[frida.core.Session]:
        if not process_name and not package_name:
            app = self._device.get_frontmost_application()
            if app:
                return self._device.attach(app.pid)
            else:
                raise ValueError("No frontmost application")

        if process_name:
            return self._device.attach(process_name)

        if package_name and spawn:
            pid = self._device.spawn([package_name])
            self._device.resume(pid)
            time.sleep(1)
            return self._device.attach(pid)

        apps = self._device.enumerate_applications(scope="full")
        for app in apps:
            if app.identifier == package_name and app.pid:
                return self._device.attach(app.pid)
        raise ValueError("Application process not found")

    def _mem_read(self, address: int, count: int) -> bytes:
        try:
            array = self._api.read_byte_array(address, count)
        except:
            array = None
        if not array:
            raise MemoryError(f"Address {hex(address)} not readable")
        return array

    def _mem_write(self, address: int, buffer: bytes) -> int:
        try:
            ret = self._api.write_byte_array(address, list(buffer))
        except:
            ret = None
        if not ret:
            raise MemoryError(f"Address {hex(address)} not writeable")
        return len(buffer)

    def _enumerate_modules(self) -> List[dict]:
        return self._api.enumerate_modules()

    def mem_alloc(self, alloc_size: int = -1, protection: str = "rwx") -> int:
        """ Allocate memory. """
        memory = int(self._api.mem_alloc(alloc_size), 16)
        self.mem_set_protect(memory, alloc_size, protection)
        return memory & 0xFFFFFFFFFFFF

    def mem_free(self, address: int) -> None:
        """ Free allocated memory. """
        self._api.mem_free(str(address))

    def mem_set_protect(self, address: int, length: int, new_protect: str = "rwx") -> None:
        """ Set memory protection. """
        self._api.mem_protect(address, length, new_protect)

    @_register_mem_read
    def mem_read_bool(self, address: int) -> bool:
        return self._api.read_bool(address)

    @_register_mem_read
    def mem_read_char(self, address: int) -> int:
        return self._api.read_char(address)

    @_register_mem_read
    def mem_read_uchar(self, address: int) -> int:
        return self._api.read_uchar(address)

    @_register_mem_read
    def mem_read_int16(self, address: int) -> int:
        return self._api.read_int16(address)

    @_register_mem_read
    def mem_read_uint16(self, address: int) -> int:
        return self._api.read_uint16(address)

    @_register_mem_read
    def mem_read_int32(self, address: int) -> int:
        return self._api.read_int32(address)

    @_register_mem_read
    def mem_read_uint32(self, address: int) -> int:
        return self._api.read_uint32(address)

    @_register_mem_read
    def mem_read_int64(self, address: int) -> int:
        return struct.unpack("q", self._mem_read(address, 8))[0]

    @_register_mem_read
    def mem_read_uint64(self, address: int) -> int:
        return struct.unpack("Q", self._mem_read(address, 8))[0]

    @_register_mem_read
    def mem_read_float(self, address: int) -> float:
        return self._api.read_float(address)

    @_register_mem_read
    def mem_read_double(self, address: int) -> float:
        return self._api.read_double(address)

    @_register_mem_read
    def mem_read_pointer(self, address: int) -> int:
        return int(self._api.read_pointer(address), 16) & 0xFFFFFFFFFFFF

    @_register_mem_read
    def mem_read_c_string(self, address: int) -> str:
        return self._api.read_c_string(address)

    @_register_mem_write(clazz=bool)
    def mem_write_bool(self, address: int, value: bool) -> bool:
        return self._api.write_bool(address, 1 if value else 0)

    @_register_mem_write(clazz=int)
    def mem_write_char(self, address: int, value: int) -> bool:
        return self._api.write_char(address, value)

    @_register_mem_write(clazz=int)
    def mem_write_uchar(self, address: int, value: int) -> bool:
        return self._api.write_uchar(address, value)

    @_register_mem_write(clazz=int)
    def mem_write_int16(self, address: int, value: int) -> bool:
        return self._api.write_int16(address, value)

    @_register_mem_write(clazz=int)
    def mem_write_uint16(self, address: int, value: int) -> bool:
        return self._api.write_uint16(address, value)

    @_register_mem_write(clazz=int)
    def mem_write_int32(self, address: int, value: int) -> bool:
        return self._api.write_int32(address, value)

    @_register_mem_write(clazz=int)
    def mem_write_uint32(self, address: int, value: int) -> bool:
        return self._api.write_uint32(address, value)

    def mem_write_int64(self, address: int, value: int) -> bool:
        if not isinstance(address, int):
            raise TypeError("address should be int")
        if not isinstance(value, int):
            raise TypeError(f"value should be int")
        return self._mem_write(address, struct.pack("q", value)) == 8

    def mem_write_uint64(self, address: int, value: int) -> bool:
        if not isinstance(address, int):
            raise TypeError("address should be int")
        if not isinstance(value, int):
            raise TypeError(f"value should be int")
        return self._mem_write(address, struct.pack("Q", value)) == 8

    @_register_mem_write(clazz=float)
    def mem_write_float(self, address: int, value: float) -> bool:
        return self._api.write_float(address, value)

    @_register_mem_write(clazz=float)
    def mem_write_double(self, address: int, value: float) -> bool:
        return self._api.write_double(address, value)

    @_register_mem_write(clazz=int)
    def mem_write_pointer(self, address: int, value: int) -> int:
        return self._api.write_pointer(address, value & 0xFFFFFFFFFFFF)

    @_register_mem_write(clazz=str)
    def mem_write_c_string(self, address: int, value: str) -> bool:
        return self._api.write_utf8_string(address, value)

    def get_module_base(self, module_name: str) -> int:
        if not isinstance(module_name, str):
            raise TypeError("module_name should be str")
        address = self._api.find_base_address(module_name)
        if isinstance(address, str):
            return int(address, 16)
        return 0

    def get_proc_address(self, module_name: str, proc_name: str) -> int:
        """ Get module exported function address in the injected process. """
        if not isinstance(proc_name, str):
            raise TypeError("proc_name should be str")
        if not isinstance(module_name, str):
            raise TypeError("module_name should be str")

        address = self._api.find_export_by_name(module_name, proc_name)
        if isinstance(address, str):
            return int(address, 16)
        return 0

    def call_native_function(self, func_address: int, args: Tuple[int], ret_type: int,
                             call_type: int, user_data: Optional[Any] = None) -> Any:
        if not isinstance(func_address, int):
            raise TypeError("func_address should be int")
        if not isinstance(args, tuple):
            raise TypeError("args should be a tuple of int")
        if not isinstance(ret_type, int):
            raise TypeError("ret_type should be int")
        if func_address <= 0:
            raise ValueError("func_address should be positive")
        if ret_type not in range(TYPE_CS_STRING + 1):
            raise ValueError(f"Unsupported return type")

        real_ret_type = frida_type_map.get(ret_type, "void")

        if len(args) <= 5:
            if user_data is None:
                user_data = [0, 0, 0]
            if len(args) == 0:
                ret_val = self._api.call_native_func(func_address, real_ret_type, *user_data)
            else:
                ret_val = self._api.call_native_func(func_address, real_ret_type, *user_data, *args)
        else:
            raise NotImplementedError("Too many parameters (more than 5)")

        if ret_type == TYPE_BOOL:
            real_ret_val = ret_val != 0
        elif ret_type == TYPE_VOID_P:
            real_ret_val = int(ret_val, 16) & 0xFFFFFFFFFFFF
        elif ret_type == TYPE_CHAR_P:
            real_ret_val = self.mem_read_c_string(int(ret_val, 16) & 0xFFFFFFFFFFFF)
        elif ret_type == TYPE_CS_STRING:
            real_ret_val = self._read_cs_string(ret_val & 0xFFFFFFFFFFFF)
        else:
            real_ret_val = ret_val
        return real_ret_val

    def code_nop(self, address: int, size: int) -> NativePatch:
        real_size = (size + 3) // 4
        if self.bit_long == 32:
            nop = b"\x00\x00\x80\xE2"
        else:
            nop = b"\xE0\x03\x00\xAA"
        return self.code_patch(nop * real_size, address)

    def execute_js(self, js_code: str, on_message_callback: Optional[Callable]) -> frida.core.Script:
        script = self._session.create_script(js_code)
        if on_message_callback:
            script.on("message", on_message_callback)
        script.load()
        return script



import os
import struct
from typing import *
from pathlib import Path
from functools import wraps

from uniref.define.struct import *
from uniref.define.constant import *
from uniref.util.winapi import WinApi
from uniref.util.injector import WinInjector


def _require_process_open(func):
    @wraps(func)
    def inner(*args, **kwargs):
        args[0]._open_process()
        return func(*args, **kwargs)
    return inner


class _PyMemoryScanner:

    def __init__(self, pid: int) -> None:
        self._process_id = pid

        self._bit_long = 0
        self._h_process = 0

    def __del__(self):
        if self._h_process:
            WinApi.CloseHandle(self._h_process)
        self._h_process = 0

    def _open_process(self):
        if not self._h_process:
            h_process = WinApi.OpenProcess(PROCESS_ALL_ACCESS, False, self._process_id)
            self._bit_long = WinApi.ProcessBitLong(h_process)
            self._h_process = h_process

    def _readable(self, protect: int) -> bool:
        flags = (PAGE_NOACCESS, PAGE_EXECUTE, PAGE_NOCACHE, PAGE_GUARD)
        return not any([protect & flag for flag in flags])

    @_require_process_open
    def _match(self, pattern: bytes, va_start: int, va_end: int) -> List[int]:
        found = []
        address = va_start
        mbi = MEMORY_BASIC_INFORMATION64()
        while address <= va_end:
            try:
                WinApi.VirtualQueryEx(self._h_process, address, byref(mbi), sizeof(mbi))
            except SystemError:
                address += PAGE_SIZE
                continue
            region_size = mbi.RegionSize
            if region_size == 0:
                address += PAGE_SIZE
                continue
            if not mbi.Protect or not self._readable(mbi.Protect):
                address += region_size
                continue

            n_read = c_int(0)
            buffer = (c_char * region_size)()
            try:
                status = WinApi.ReadProcessMemory(self._h_process, address, buffer, region_size, byref(n_read))
            except:
                status = False
            if status and n_read.value:
                pos = 0
                bytes_read = buffer.raw
                pattern_len = len(pattern)
                while True:
                    idx = bytes_read.find(pattern)
                    if idx == -1:
                        break
                    pos += idx + pattern_len
                    found.append(address + pos - pattern_len)
                    bytes_read = bytes_read[pos:]
            address += region_size
        return found

    def scan_uint64(self, value: int, va_start: int = 0, va_end: int = -1) -> List[int]:
        if va_end <= 0:
            va_end = 0xffffffff if self._bit_long == 32 else 0x7fffffffffff
        if va_start > va_end:
            raise ValueError("va_start > va_end")
        return self._match(struct.pack("Q", value), va_start, va_end)


class CMemoryScanner:

    def __init__(self, pid: int, bit_long: int) -> None:
        self._process_id = pid
        self._bit_long = bit_long
        self._injector = WinInjector()
        self._init_proc()
    
    def _init_proc(self):
        _util = cdll.LoadLibrary(str(Path(os.path.abspath(__file__)).parent.parent / "bin/win/util64.dll"))

        self._SearchChar = _util.SearchChar
        self._SearchChar.argtypes = [c_uint32, c_byte, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchChar.restype = c_uint64

        self._SearchUChar = _util.SearchUChar
        self._SearchUChar.argtypes = [c_uint32, c_ubyte, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchUChar.restype = c_uint64

        self._SearchInt16 = _util.SearchInt16
        self._SearchInt16.argtypes = [c_uint32, c_int16, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchInt16.restype = c_uint64

        self._SearchUInt16 = _util.SearchUInt16
        self._SearchUInt16.argtypes = [c_uint32, c_uint16, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchUInt16.restype = c_uint64

        self._SearchInt32 = _util.SearchInt32
        self._SearchInt32.argtypes = [c_uint32, c_int32, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchInt32.restype = c_uint64

        self._SearchUInt32 = _util.SearchUInt32
        self._SearchUInt32.argtypes = [c_uint32, c_uint32, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchUInt32.restype = c_uint64

        self._SearchInt64 = _util.SearchInt64
        self._SearchInt64.argtypes = [c_uint32, c_int64, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchInt64.restype = c_uint64

        self._SearchUInt64 = _util.SearchUInt64
        self._SearchUInt64.argtypes = [c_uint32, c_uint64, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchUInt64.restype = c_uint64

        self._SearchFloat = _util.SearchFloat
        self._SearchFloat.argtypes = [c_uint32, c_float, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchFloat.restype = c_uint64

        self._SearchDouble = _util.SearchDouble
        self._SearchDouble.argtypes = [c_uint32, c_double, c_bool, c_uint64, c_uint64, c_void_p]
        self._SearchDouble.restype = c_uint64

        self._ResFree = _util.ResFree
        self._ResFree.argtypes = [c_void_p]

        self._util = _util

    def _res_free(self, ptr: int):
        self._ResFree(ptr)

    def _do_scan(self, elem_type: int, target: Any, writeable: bool, va_start: int = 0, va_end: int = -1):
        if va_end <= 0:
            va_end = 0xffffffff if self._bit_long == 32 else 0x7fffffffffff
        if va_start < 0:
            raise ValueError("va_start < 0")
        if va_start > va_end:
            raise ValueError("va_start > va_end")

        if elem_type == TYPE_CHAR:
            scan_func = self._SearchChar
        elif elem_type == TYPE_UCHAR:
            scan_func = self._SearchUChar
        elif elem_type == TYPE_INT16:
            scan_func = self._SearchInt16
        elif elem_type == TYPE_UINT16:
            scan_func = self._SearchUInt16
        elif elem_type == TYPE_INT32:
            scan_func = self._SearchInt32
        elif elem_type == TYPE_UINT32:
            scan_func = self._SearchUInt32
        elif elem_type == TYPE_INT64:
            scan_func = self._SearchInt64
        elif elem_type == TYPE_UINT64:
            scan_func = self._SearchUInt64
        elif elem_type == TYPE_FLOAT:
            scan_func = self._SearchFloat
        elif elem_type == TYPE_DOUBLE:
            scan_func = self._SearchDouble
        else:
            raise ValueError("Element type unsupported")
        res = c_void_p()
        cnt = scan_func(self._process_id, target, writeable, va_start, va_end, byref(res))
        if cnt:
            found = self._injector.mem_read_pointer_array(res.value, cnt)
            self._res_free(res.value)
            return found
        return []

    def scan_char(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        return self._do_scan(TYPE_CHAR, target, writeable, va_start, va_end)

    def scan_uchar(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        return self._do_scan(TYPE_UCHAR, target, writeable, va_start, va_end)

    def scan_int16(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        return self._do_scan(TYPE_INT16, target, writeable, va_start, va_end)

    def scan_uint16(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        return self._do_scan(TYPE_UINT16, target, writeable, va_start, va_end)

    def scan_int32(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        return self._do_scan(TYPE_INT32, target, writeable, va_start, va_end)

    def scan_uint32(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        return self._do_scan(TYPE_UINT32, target, writeable, va_start, va_end)

    def scan_int64(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        return self._do_scan(TYPE_INT64, target, writeable, va_start, va_end)

    def scan_uint64(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        return self._do_scan(TYPE_UINT64, target, writeable, va_start, va_end)

    def scan_pointer(self, target: int, writeable: bool, va_start: int = 0, va_end: int = -1):
        if self._bit_long == 32:
            return self.scan_uint32(target, writeable, va_start, va_end)
        else:
            return self.scan_uint64(target, writeable, va_start, va_end)



class NativePatch:
    """ This class's instance is returned by calling ``MonoMethod.native_patch`` and ``MonoMethod.native_nop``.
    You can use it to manage the patch. """

    def __init__(self, injector: object, address: int, enabled_code: bytes, disabled_code: bytes) -> None:
        self._address = address
        self._injector = injector
        self._enabled_code = enabled_code
        self._disabled_code = disabled_code

        self._enable = False
        self._try_patch()

    def _try_patch(self):
        disabled_code_len = len(self._disabled_code)
        while len(self._enabled_code) < disabled_code_len:
            self._enabled_code += b'\x90'

        if len(self._enabled_code) > disabled_code_len:
            raise NotImplementedError
        self.enable()

    @property
    def address(self) -> int:
        """ patch address """
        return self._address

    def is_enabled(self) -> bool:
        """ Check if the patch is enabled. """
        return self._enable

    def enable(self) -> None:
        """ Enable the patch. """
        if self._enable:
            return

        enabled_code_len = len(self._enabled_code)

        old_protect = self._injector.mem_set_protect(self._address, enabled_code_len)
        self._injector.mem_write_bytes(self._address, self._enabled_code)
        if isinstance(old_protect, str):
            self._injector.mem_set_protect(self._address, enabled_code_len, old_protect)

        self._enable = True

    def disable(self) -> None:
        """ Disable the patch. """
        if not self._enable:
            return

        disabled_code_len = len(self._disabled_code)

        old_protect = self._injector.mem_set_protect(self._address, disabled_code_len)
        self._injector.mem_write_bytes(self._address, self._disabled_code)
        if isinstance(old_protect, str):
            self._injector.mem_set_protect(self._address, disabled_code_len, old_protect)

        self._enable = False
